######################################################################
#
# File: b2sdk/_internal/utils/typing.py
#
# Copyright 2023 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from __future__ import annotations

from typing import Dict, List, Union

try:
    from typing_extensions import TypeAlias
except ImportError:
    from typing import TypeAlias

JSON: TypeAlias = Union[Dict[str, 'JSON'], List['JSON'], str, int, float, bool, None]



######################################################################
#
# File: b2sdk/v2/session.py
#
# Copyright 2021 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from __future__ import annotations

from b2sdk import v3
from .b2http import B2Http

from .account_info import SqliteAccountInfo
from ._compat import _file_infos_rename
from . import api_config as _api_config
from .._internal import cache as _cache
from .._internal.account_info import abstract as _abstract


# Override to use legacy B2Http
class B2Session(v3.B2Session):
    SQLITE_ACCOUNT_INFO_CLASS = staticmethod(SqliteAccountInfo)
    B2HTTP_CLASS = staticmethod(B2Http)

    def __init__(
        self,
        account_info: _abstract.AbstractAccountInfo | None = None,
        cache: _cache.AbstractCache | None = None,
        api_config: _api_config.B2HttpApiConfig = _api_config.DEFAULT_HTTP_API_CONFIG,
    ):
        if account_info is not None and cache is None:
            # preserve legacy behavior https://github.com/Backblaze/b2-sdk-python/issues/497#issuecomment-2147461352
            cache = _cache.DummyCache()

        super().__init__(account_info, cache, api_config)

    def create_key(
        self, account_id, capabilities, key_name, valid_duration_seconds, bucket_id, name_prefix
    ):
        return self._wrap_default_token(
            self.raw_api.create_key,
            account_id,
            capabilities,
            key_name,
            valid_duration_seconds,
            bucket_id,
            name_prefix,
        )

    @_file_infos_rename
    def upload_file(
        self,
        bucket_id,
        file_name,
        content_length,
        content_type,
        content_sha1,
        file_info,
        data_stream,
        server_side_encryption: v3.EncryptionSetting | None = None,
        file_retention: v3.FileRetentionSetting | None = None,
        legal_hold: v3.LegalHold | None = None,
        custom_upload_timestamp: int | None = None,
        cache_control: str | None = None,
        *args,
        **kwargs,
    ):
        if cache_control is not None:
            file_info['b2-cache-control'] = cache_control
        return super().upload_file(
            bucket_id,
            file_name,
            content_length,
            content_type,
            content_sha1,
            file_info,
            data_stream,
            server_side_encryption,
            file_retention,
            legal_hold,
            custom_upload_timestamp,
            *args,
            **kwargs,
        )

    def _construct_allowed_dict(self, storage_api_info):
        return {
            'bucketId': storage_api_info['bucketId'],
            'bucketName': storage_api_info['bucketName'],
            'capabilities': storage_api_info['capabilities'],
            'namePrefix': storage_api_info['namePrefix'],
        }

    def _get_allowed_buckets_message(self, allowed) -> str | None:
        bucket_name = allowed['bucketName']
        if bucket_name is None:
            return None

        return "restricted to bucket '" + bucket_name + "'"



######################################################################
#
# File: b2sdk/_internal/file_lock.py
#
# Copyright 2021 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from __future__ import annotations

import enum

from .exception import UnexpectedCloudBehaviour

ACTIONS_WITHOUT_LOCK_SETTINGS = frozenset(['hide', 'folder'])


@enum.unique
class RetentionMode(enum.Enum):
    """Enum class representing retention modes set in files and buckets"""

    GOVERNANCE = 'governance'  #: retention settings for files in this mode can be modified by clients with appropriate application key capabilities
    COMPLIANCE = 'compliance'  #: retention settings for files in this mode can only be modified by extending the retention dates by clients with appropriate application key capabilities
    NONE = None  #: retention not set
    UNKNOWN = 'unknown'  #: the client is not authorized to read retention settings


RETENTION_MODES_REQUIRING_PERIODS = frozenset({RetentionMode.COMPLIANCE, RetentionMode.GOVERNANCE})


class RetentionPeriod:
    """Represent a time period (either in days or in years) that is used as a default for bucket retention"""

    KNOWN_UNITS = ['days', 'years']

    def __init__(self, years: int | None = None, days: int | None = None):
        """Create a retention period, provide exactly one of: days, years"""
        assert (years is None) != (days is None)
        if years is not None:
            self.duration = years
            self.unit = 'years'
        else:
            self.duration = days
            self.unit = 'days'

    @classmethod
    def from_period_dict(cls, period_dict):
        """
        Build a RetentionPeriod from an object returned by the server, such as:

        .. code-block ::

            {
                "duration": 2,
                "unit": "years"
            }
        """
        assert period_dict['unit'] in cls.KNOWN_UNITS
        return cls(**{period_dict['unit']: period_dict['duration']})

    def as_dict(self):
        return {
            'duration': self.duration,
            'unit': self.unit,
        }

    def __repr__(self):
        return f'{self.__class__.__name__}({self.duration} {self.unit})'

    def __eq__(self, other):
        return self.unit == other.unit and self.duration == other.duration


class FileRetentionSetting:
    """
    Represent file retention settings, i.e. whether the file is retained, in which mode and until when

    :param mode: retention mode
    :type mode: RetentionMode
    :param retain_until: retain until timestamp (in milliseconds since :abbr:`epoch (1970-01-01 00:00:00)`)
    :type retain_until: int
    """

    def __init__(self, mode: RetentionMode, retain_until: int | None = None):
        """
        :param mode: retention mode
        :type mode: RetentionMode
        :param retain_until: retain until timestamp (in milliseconds since :abbr:`epoch (1970-01-01 00:00:00)`)
        :type retain_until: int
        """
        if mode in RETENTION_MODES_REQUIRING_PERIODS and retain_until is None:
            raise ValueError(f'must specify retain_until for retention mode {mode}')
        self.mode = mode
        self.retain_until = retain_until

    @classmethod
    def from_file_version_dict(cls, file_version_dict: dict) -> FileRetentionSetting:
        """
        Returns FileRetentionSetting for the given file_version_dict retrieved from the api. E.g.

        .. code-block ::

            {
                "action": "upload",
                "fileRetention": {
                    "isClientAuthorizedToRead": false,
                    "value": null
                },
                ...
            }

            {
                "action": "upload",
                "fileRetention": {
                    "isClientAuthorizedToRead": true,
                    "value": {
                      "mode": "governance",
                      "retainUntilTimestamp": 1628942493000
                    }
                },
                ...
            }
        """
        if 'fileRetention' not in file_version_dict:
            if file_version_dict['action'] not in ACTIONS_WITHOUT_LOCK_SETTINGS:
                raise UnexpectedCloudBehaviour(
                    'No fileRetention provided for file version with action=%s'
                    % (file_version_dict['action'])
                )
            return NO_RETENTION_FILE_SETTING
        file_retention_dict = file_version_dict['fileRetention']

        if not file_retention_dict['isClientAuthorizedToRead']:
            return cls(RetentionMode.UNKNOWN, None)

        return cls.from_file_retention_value_dict(file_retention_dict['value'])

    @classmethod
    def from_file_retention_value_dict(
        cls, file_retention_value_dict: dict
    ) -> FileRetentionSetting:
        mode = file_retention_value_dict['mode']
        if mode is None:
            return NO_RETENTION_FILE_SETTING

        return cls(
            RetentionMode(mode),
            file_retention_value_dict['retainUntilTimestamp'],
        )

    @classmethod
    def from_server_response(cls, server_response: dict) -> FileRetentionSetting:
        return cls.from_file_retention_value_dict(server_response['fileRetention'])

    @classmethod
    def from_response_headers(cls, headers) -> FileRetentionSetting:
        retention_mode_header = 'X-Bz-File-Retention-Mode'
        retain_until_header = 'X-Bz-File-Retention-Retain-Until-Timestamp'
        if retention_mode_header in headers:
            if retain_until_header in headers:
                retain_until = int(headers[retain_until_header])
            else:
                retain_until = None
            return cls(RetentionMode(headers[retention_mode_header]), retain_until)
        if 'X-Bz-Client-Unauthorized-To-Read' in headers and retention_mode_header in headers[
            'X-Bz-Client-Unauthorized-To-Read'
        ].split(','):
            return UNKNOWN_FILE_RETENTION_SETTING
        return NO_RETENTION_FILE_SETTING  # the bucket is not file-lock-enabled or the file is has no retention set

    def serialize_to_json_for_request(self):
        if self.mode is RetentionMode.UNKNOWN:
            raise ValueError('cannot use an unknown file retention setting in requests')
        return self.as_dict()

    def as_dict(self):
        return {
            'mode': self.mode.value,
            'retainUntilTimestamp': self.retain_until,
        }

    def add_to_to_upload_headers(self, headers):
        if self.mode is RetentionMode.UNKNOWN:
            raise ValueError('cannot use an unknown file retention setting in requests')

        headers['X-Bz-File-Retention-Mode'] = str(
            self.mode.value
        )  # mode = NONE is not supported by the server at the
        # moment, but it should be
        headers['X-Bz-File-Retention-Retain-Until-Timestamp'] = str(self.retain_until)

    def __eq__(self, other):
        return self.mode == other.mode and self.retain_until == other.retain_until

    def __repr__(self):
        return f'{self.__class__.__name__}({repr(self.mode.value)}, {repr(self.retain_until)})'


@enum.unique
class LegalHold(enum.Enum):
    """Enum holding information about legalHold switch in a file."""

    ON = 'on'  #: legal hold set to "on"
    OFF = 'off'  #: legal hold set to "off"
    UNSET = None  #: server default, as for now it is functionally equivalent to OFF
    UNKNOWN = 'unknown'  #: the client is not authorized to read legal hold settings

    def is_on(self):
        """Is the legalHold switch on?"""
        return self is LegalHold.ON

    def is_off(self):
        """Is the legalHold switch off or left as default (which also means off)?"""
        return self is LegalHold.OFF or self is LegalHold.UNSET

    def is_unknown(self):
        """Is the legalHold switch unknown?"""
        return self is LegalHold.UNKNOWN

    @classmethod
    def from_file_version_dict(cls, file_version_dict: dict) -> LegalHold:
        if 'legalHold' not in file_version_dict:
            if file_version_dict['action'] not in ACTIONS_WITHOUT_LOCK_SETTINGS:
                raise UnexpectedCloudBehaviour(
                    'legalHold not provided for file version with action=%s'
                    % (file_version_dict['action'])
                )
            return cls.UNSET
        if not file_version_dict['legalHold']['isClientAuthorizedToRead']:
            return cls.UNKNOWN
        return cls.from_string_or_none(file_version_dict['legalHold']['value'])

    @classmethod
    def from_server_response(cls, server_response: dict) -> LegalHold:
        return cls.from_string_or_none(server_response['legalHold'])

    @classmethod
    def from_string_or_none(cls, string: str | None) -> LegalHold:
        return cls(string)

    @classmethod
    def from_response_headers(cls, headers) -> LegalHold:
        legal_hold_header = 'X-Bz-File-Legal-Hold'
        if legal_hold_header in headers:
            return cls(headers['X-Bz-File-Legal-Hold'])
        if 'X-Bz-Client-Unauthorized-To-Read' in headers and legal_hold_header in headers[
            'X-Bz-Client-Unauthorized-To-Read'
        ].split(','):
            return cls.UNKNOWN
        return (
            cls.UNSET
        )  # the bucket is not file-lock-enabled or the header is missing for any other reason

    def to_server(self) -> str:
        if self.is_unknown():
            raise ValueError('Cannot use an unknown legal hold in requests')
        if self.is_on():
            return self.__class__.ON.value
        return self.__class__.OFF.value

    def add_to_upload_headers(self, headers):
        headers['X-Bz-File-Legal-Hold'] = self.to_server()


class BucketRetentionSetting:
    """Represent bucket's default file retention settings, i.e. whether the files should be retained, in which mode
    and for how long"""

    def __init__(self, mode: RetentionMode, period: RetentionPeriod | None = None):
        if mode in RETENTION_MODES_REQUIRING_PERIODS and period is None:
            raise ValueError(f'must specify period for retention mode {mode}')
        self.mode = mode
        self.period = period

    @classmethod
    def from_bucket_retention_dict(cls, retention_dict: dict):
        """
        Build a BucketRetentionSetting from an object returned by the server, such as:

        .. code-block::

            {
                "mode": "compliance",
                "period": {
                    "duration": 7,
                    "unit": "days"
                }
            }

        """
        period = retention_dict['period']
        if period is not None:
            period = RetentionPeriod.from_period_dict(period)
        return cls(RetentionMode(retention_dict['mode']), period)

    def as_dict(self):
        result = {
            'mode': self.mode.value,
        }
        if self.period is not None:
            result['period'] = self.period.as_dict()
        return result

    def serialize_to_json_for_request(self):
        if self.mode == RetentionMode.UNKNOWN:
            raise ValueError('cannot use an unknown file lock configuration in requests')
        return self.as_dict()

    def __eq__(self, other):
        return self.mode == other.mode and self.period == other.period

    def __repr__(self):
        return f'{self.__class__.__name__}({repr(self.mode.value)}, {repr(self.period)})'


class FileLockConfiguration:
    """Represent bucket's file lock configuration, i.e. whether the file lock mechanism is enabled and default
    file retention"""

    def __init__(
        self,
        default_retention: BucketRetentionSetting,
        is_file_lock_enabled: bool | None,
    ):
        self.default_retention = default_retention
        self.is_file_lock_enabled = is_file_lock_enabled

    @classmethod
    def from_bucket_dict(cls, bucket_dict):
        """
        Build a FileLockConfiguration from an object returned by server, such as:

        .. code-block::

            {
                "isClientAuthorizedToRead": true,
                "value": {
                    "defaultRetention": {
                        "mode": "governance",
                        "period": {
                            "duration": 2,
                            "unit": "years"
                        }
                    },
                    "isFileLockEnabled": true
                }
            }

            or

            {
                "isClientAuthorizedToRead": false,
                "value": null
            }
        """

        if not bucket_dict['fileLockConfiguration']['isClientAuthorizedToRead']:
            return cls(UNKNOWN_BUCKET_RETENTION, None)
        retention = BucketRetentionSetting.from_bucket_retention_dict(
            bucket_dict['fileLockConfiguration']['value']['defaultRetention']
        )
        is_file_lock_enabled = bucket_dict['fileLockConfiguration']['value']['isFileLockEnabled']
        return cls(retention, is_file_lock_enabled)

    def as_dict(self):
        return {
            'defaultRetention': self.default_retention.as_dict(),
            'isFileLockEnabled': self.is_file_lock_enabled,
        }

    def __eq__(self, other):
        return (
            self.default_retention == other.default_retention
            and self.is_file_lock_enabled == other.is_file_lock_enabled
        )

    def __repr__(self):
        return f'{self.__class__.__name__}({repr(self.default_retention)}, {repr(self.is_file_lock_enabled)})'


UNKNOWN_BUCKET_RETENTION = BucketRetentionSetting(RetentionMode.UNKNOWN)
"""Commonly used "unknown" default bucket retention setting"""
UNKNOWN_FILE_LOCK_CONFIGURATION = FileLockConfiguration(UNKNOWN_BUCKET_RETENTION, None)
"""Commonly used "unknown" bucket file lock setting"""
NO_RETENTION_BUCKET_SETTING = BucketRetentionSetting(RetentionMode.NONE)
"""Commonly used "no retention" default bucket retention"""
NO_RETENTION_FILE_SETTING = FileRetentionSetting(RetentionMode.NONE)
"""Commonly used "no retention" file setting"""
UNKNOWN_FILE_RETENTION_SETTING = FileRetentionSetting(RetentionMode.UNKNOWN)
"""Commonly used "unknown" file retention setting"""



######################################################################
#
# File: b2sdk/v2/replication/setup.py
#
# Copyright 2025 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from __future__ import annotations

from b2sdk._internal.replication.setup import *  # noqa

from b2sdk import v3 as v3

from ..bucket import Bucket

from ..application_key import ApplicationKey  # type: ignore


class ReplicationSetupHelper(v3.ReplicationSetupHelper):  # type: ignore
    def setup_source(
        self,
        source_bucket: Bucket,
        source_key: ApplicationKey,
        destination_bucket: Bucket,
        prefix: str | None = None,
        name: str | None = None,
        priority: int = None,
        include_existing_files: bool = False,
    ) -> Bucket:
        return super().setup_source(
            source_bucket=source_bucket,
            source_key=source_key,
            destination_bucket=destination_bucket,
            prefix=prefix,
            name=name,
            priority=priority,
            include_existing_files=include_existing_files,
        )

    @classmethod
    def _get_source_key(
        cls,
        source_bucket: Bucket,
        prefix: str,
        current_replication_configuration: ReplicationConfiguration,
    ) -> ApplicationKey:
        return super()._get_source_key(
            source_bucket=source_bucket,
            prefix=prefix,
            current_replication_configuration=current_replication_configuration,
        )

    @classmethod
    def _should_make_new_source_key(
        cls,
        current_replication_configuration: ReplicationConfiguration,
        current_source_key: ApplicationKey | None,
    ) -> bool:
        return super()._should_make_new_source_key(
            current_replication_configuration=current_replication_configuration,
            current_source_key=current_source_key,
        )

    @classmethod
    def _create_source_key(
        cls,
        name: str,
        bucket: Bucket,
        prefix: str | None = None,
    ) -> ApplicationKey:
        return super()._create_source_key(
            name=name,
            bucket=bucket,
            prefix=prefix,
        )

    @classmethod
    def _create_destination_key(
        cls,
        name: str,
        bucket: Bucket,
        prefix: str | None = None,
    ) -> ApplicationKey:
        return super()._create_destination_key(
            name=name,
            bucket=bucket,
            prefix=prefix,
        )

    @classmethod
    def _create_key(
        cls,
        name: str,
        bucket: Bucket,
        prefix: str | None = None,
        capabilities=tuple(),
    ) -> ApplicationKey:
        api: B2Api = bucket.api
        return api.create_key(
            capabilities=capabilities,
            key_name=name,
            bucket_id=bucket.id_,
            name_prefix=prefix,
        )



######################################################################
#
# File: b2sdk/v3/__init__.py
#
# Copyright 2021 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from __future__ import annotations

# Set default logging handler to avoid "No handler found" warnings.
import logging as _logging

_logging.getLogger('b2sdk').addHandler(_logging.NullHandler())


class UrllibWarningFilter:
    def filter(self, record):
        return record.msg != 'Connection pool is full, discarding connection: %s'


_logging.getLogger('urllib3.connectionpool').addFilter(UrllibWarningFilter())

# this file maps the external interface into internal interface
# it will come handy if we ever need to move something

# core

from b2sdk._internal.api import B2Api
from b2sdk._internal.api import Services
from b2sdk._internal.bucket import Bucket
from b2sdk._internal.bucket import BucketFactory
from b2sdk._internal.raw_api import ALL_CAPABILITIES, REALM_URLS, EVENT_TYPE

# encryption

from b2sdk._internal.encryption.setting import EncryptionSetting
from b2sdk._internal.encryption.setting import EncryptionSettingFactory
from b2sdk._internal.encryption.setting import EncryptionKey
from b2sdk._internal.encryption.setting import SSE_NONE, SSE_B2_AES, UNKNOWN_KEY_ID
from b2sdk._internal.encryption.types import EncryptionAlgorithm
from b2sdk._internal.encryption.types import EncryptionMode
from b2sdk._internal.http_constants import SSE_C_KEY_ID_FILE_INFO_KEY_NAME

# account info

from b2sdk._internal.account_info.abstract import AbstractAccountInfo
from b2sdk._internal.account_info.in_memory import InMemoryAccountInfo
from b2sdk._internal.account_info.sqlite_account_info import SqliteAccountInfo
from b2sdk._internal.account_info.sqlite_account_info import B2_ACCOUNT_INFO_ENV_VAR
from b2sdk._internal.account_info.sqlite_account_info import B2_ACCOUNT_INFO_DEFAULT_FILE
from b2sdk._internal.account_info.sqlite_account_info import B2_ACCOUNT_INFO_PROFILE_FILE
from b2sdk._internal.account_info.sqlite_account_info import XDG_CONFIG_HOME_ENV_VAR
from b2sdk._internal.account_info.stub import StubAccountInfo
from b2sdk._internal.account_info.upload_url_pool import UploadUrlPool
from b2sdk._internal.account_info.upload_url_pool import UrlPoolAccountInfo

# version & version utils

from b2sdk.version import VERSION, USER_AGENT
from b2sdk._internal.version_utils import rename_argument, rename_function, FeaturePreviewWarning

# utils

from b2sdk._internal.utils import (
    b2_url_encode,
    b2_url_decode,
    choose_part_ranges,
    current_time_millis,
    fix_windows_path_limit,
    format_and_scale_fraction,
    format_and_scale_number,
    hex_sha1_of_stream,
    hex_sha1_of_bytes,
    hex_sha1_of_file,
    IncrementalHexDigester,
)

from b2sdk._internal.utils.filesystem import (
    points_to_fifo,
    points_to_stdout,
    STDOUT_FILEPATH,
)
from b2sdk._internal.utils import trace_call
from b2sdk._internal.utils.docs import get_b2sdk_doc_urls

# data classes

from b2sdk._internal.application_key import ApplicationKey
from b2sdk._internal.application_key import BaseApplicationKey
from b2sdk._internal.application_key import FullApplicationKey
from b2sdk._internal.file_version import DownloadVersion
from b2sdk._internal.file_version import DownloadVersionFactory
from b2sdk._internal.file_version import FileIdAndName
from b2sdk._internal.file_version import FileVersion
from b2sdk._internal.file_version import FileVersionFactory
from b2sdk._internal.large_file.part import Part
from b2sdk._internal.large_file.unfinished_large_file import UnfinishedLargeFile
from b2sdk._internal.large_file.services import LargeFileServices
from b2sdk._internal.utils.range_ import Range, EMPTY_RANGE

# file lock

from b2sdk._internal.file_lock import BucketRetentionSetting
from b2sdk._internal.file_lock import FileLockConfiguration
from b2sdk._internal.file_lock import FileRetentionSetting
from b2sdk._internal.file_lock import LegalHold
from b2sdk._internal.file_lock import NO_RETENTION_BUCKET_SETTING
from b2sdk._internal.file_lock import NO_RETENTION_FILE_SETTING
from b2sdk._internal.file_lock import RetentionMode
from b2sdk._internal.file_lock import RetentionPeriod
from b2sdk._internal.file_lock import UNKNOWN_BUCKET_RETENTION
from b2sdk._internal.file_lock import UNKNOWN_FILE_LOCK_CONFIGURATION
from b2sdk._internal.file_lock import UNKNOWN_FILE_RETENTION_SETTING

# progress reporting

from b2sdk._internal.progress import AbstractProgressListener
from b2sdk._internal.progress import DoNothingProgressListener
from b2sdk._internal.progress import ProgressListenerForTest
from b2sdk._internal.progress import SimpleProgressListener
from b2sdk._internal.progress import TqdmProgressListener
from b2sdk._internal.progress import make_progress_listener

# raw_simulator

from b2sdk._internal.raw_simulator import BucketSimulator
from b2sdk._internal.raw_simulator import FakeResponse
from b2sdk._internal.raw_simulator import FileSimulator
from b2sdk._internal.raw_simulator import KeySimulator
from b2sdk._internal.raw_simulator import PartSimulator
from b2sdk._internal.raw_simulator import RawSimulator

# raw_api

from b2sdk._internal.raw_api import AbstractRawApi
from b2sdk._internal.raw_api import B2RawHTTPApi
from b2sdk._internal.raw_api import MetadataDirectiveMode
from b2sdk._internal.raw_api import LifecycleRule
from b2sdk._internal.raw_api import (
    NotificationRule,
    NotificationRuleResponse,
    notification_rule_response_to_request,
)

# stream

from b2sdk._internal.stream.chained import StreamOpener
from b2sdk._internal.stream.progress import AbstractStreamWithProgress
from b2sdk._internal.stream import RangeOfInputStream
from b2sdk._internal.stream import ReadingStreamWithProgress
from b2sdk._internal.stream import StreamWithHash
from b2sdk._internal.stream import WritingStreamWithProgress

# source / destination

from b2sdk._internal.transfer.inbound.downloaded_file import DownloadedFile
from b2sdk._internal.transfer.inbound.downloaded_file import MtimeUpdatedFile
from b2sdk._internal.transfer.inbound.download_manager import DownloadManager

from b2sdk._internal.transfer.outbound.outbound_source import OutboundTransferSource
from b2sdk._internal.transfer.outbound.copy_source import CopySource
from b2sdk._internal.transfer.outbound.upload_source import AbstractUploadSource
from b2sdk._internal.transfer.outbound.upload_source import UploadSourceBytes
from b2sdk._internal.transfer.outbound.upload_source import UploadSourceLocalFile
from b2sdk._internal.transfer.outbound.upload_source import UploadSourceLocalFileRange
from b2sdk._internal.transfer.outbound.upload_source import UploadSourceStream
from b2sdk._internal.transfer.outbound.upload_source import UploadSourceStreamRange
from b2sdk._internal.transfer.outbound.upload_manager import UploadManager

from b2sdk._internal.transfer.emerge.planner.upload_subpart import CachedBytesStreamOpener
from b2sdk._internal.transfer.emerge.write_intent import WriteIntent

# transfer

from b2sdk._internal.transfer.inbound.downloader.abstract import AbstractDownloader
from b2sdk._internal.transfer.outbound.large_file_upload_state import LargeFileUploadState
from b2sdk._internal.transfer.inbound.downloader.parallel import ParallelDownloader
from b2sdk._internal.transfer.inbound.downloader.parallel import PartToDownload
from b2sdk._internal.transfer.inbound.downloader.parallel import WriterThread
from b2sdk._internal.transfer.outbound.progress_reporter import PartProgressReporter
from b2sdk._internal.transfer.inbound.downloader.simple import SimpleDownloader

# sync

from b2sdk._internal.sync.action import AbstractAction
from b2sdk._internal.sync.action import B2CopyAction
from b2sdk._internal.sync.action import B2DeleteAction
from b2sdk._internal.sync.action import B2DownloadAction
from b2sdk._internal.sync.action import B2HideAction
from b2sdk._internal.sync.action import B2UploadAction
from b2sdk._internal.sync.action import LocalDeleteAction
from b2sdk._internal.sync.exception import IncompleteSync
from b2sdk._internal.sync.policy import AbstractFileSyncPolicy
from b2sdk._internal.sync.policy import CompareVersionMode
from b2sdk._internal.sync.policy import NewerFileSyncMode
from b2sdk._internal.sync.policy import DownAndDeletePolicy
from b2sdk._internal.sync.policy import DownAndKeepDaysPolicy
from b2sdk._internal.sync.policy import DownPolicy
from b2sdk._internal.sync.policy import CopyPolicy
from b2sdk._internal.sync.policy import CopyAndDeletePolicy
from b2sdk._internal.sync.policy import CopyAndKeepDaysPolicy
from b2sdk._internal.sync.policy import UpAndDeletePolicy
from b2sdk._internal.sync.policy import UpAndKeepDaysPolicy
from b2sdk._internal.sync.policy import UpPolicy
from b2sdk._internal.sync.policy import make_b2_keep_days_actions
from b2sdk._internal.sync.policy_manager import SyncPolicyManager
from b2sdk._internal.sync.policy_manager import POLICY_MANAGER
from b2sdk._internal.sync.report import SyncFileReporter
from b2sdk._internal.sync.report import SyncReport
from b2sdk._internal.sync.sync import KeepOrDeleteMode
from b2sdk._internal.sync.sync import Synchronizer
from b2sdk._internal.sync.sync import UploadMode
from b2sdk._internal.sync.encryption_provider import AbstractSyncEncryptionSettingsProvider
from b2sdk._internal.sync.encryption_provider import BasicSyncEncryptionSettingsProvider
from b2sdk._internal.sync.encryption_provider import ServerDefaultSyncEncryptionSettingsProvider
from b2sdk._internal.sync.encryption_provider import (
    SERVER_DEFAULT_SYNC_ENCRYPTION_SETTINGS_PROVIDER,
)

# scan

from b2sdk._internal.scan.exception import EnvironmentEncodingError
from b2sdk._internal.scan.exception import InvalidArgument
from b2sdk._internal.scan.folder import AbstractFolder
from b2sdk._internal.scan.folder import B2Folder
from b2sdk._internal.scan.folder import LocalFolder
from b2sdk._internal.scan.folder_parser import parse_folder
from b2sdk._internal.scan.path import AbstractPath, B2Path, LocalPath
from b2sdk._internal.scan.policies import convert_dir_regex_to_dir_prefix_regex
from b2sdk._internal.scan.policies import DEFAULT_SCAN_MANAGER
from b2sdk._internal.scan.policies import IntegerRange
from b2sdk._internal.scan.policies import RegexSet
from b2sdk._internal.scan.policies import ScanPoliciesManager
from b2sdk._internal.scan.report import ProgressReport
from b2sdk._internal.scan.scan import zip_folders
from b2sdk._internal.scan.scan import AbstractScanResult
from b2sdk._internal.scan.scan import AbstractScanReport
from b2sdk._internal.scan.scan import CountAndSampleScanReport

# replication

from b2sdk._internal.replication.setting import ReplicationConfigurationFactory
from b2sdk._internal.replication.setting import ReplicationConfiguration
from b2sdk._internal.replication.setting import ReplicationRule
from b2sdk._internal.replication.types import ReplicationStatus
from b2sdk._internal.replication.setup import ReplicationSetupHelper
from b2sdk._internal.replication.monitoring import ReplicationScanResult
from b2sdk._internal.replication.monitoring import ReplicationReport
from b2sdk._internal.replication.monitoring import ReplicationMonitor

# other

from b2sdk._internal.included_sources import get_included_sources
from b2sdk._internal.b2http import B2Http
from b2sdk._internal.api_config import B2HttpApiConfig
from b2sdk._internal.api_config import DEFAULT_HTTP_API_CONFIG
from b2sdk._internal.b2http import ClockSkewHook
from b2sdk._internal.b2http import HttpCallback
from b2sdk._internal.b2http import ResponseContextManager
from b2sdk._internal.bounded_queue_executor import BoundedQueueExecutor
from b2sdk._internal.cache import AbstractCache
from b2sdk._internal.cache import AuthInfoCache
from b2sdk._internal.cache import DummyCache
from b2sdk._internal.cache import InMemoryCache
from b2sdk._internal.http_constants import (
    BUCKET_NAME_CHARS,
    BUCKET_NAME_CHARS_UNIQ,
    BUCKET_NAME_LENGTH_RANGE,
    DEFAULT_MAX_PART_SIZE,
    DEFAULT_MIN_PART_SIZE,
    DEFAULT_RECOMMENDED_UPLOAD_PART_SIZE,
    LARGE_FILE_SHA1,
    LIST_FILE_NAMES_MAX_LIMIT,
    SRC_LAST_MODIFIED_MILLIS,
)
from b2sdk._internal.session import B2Session
from b2sdk._internal.utils.thread_pool import ThreadPoolMixin
from b2sdk._internal.utils.escape import (
    unprintable_to_hex,
    escape_control_chars,
    substitute_control_chars,
)

# filter
from b2sdk._internal.filter import FilterType, Filter

# typing
from b2sdk._internal.utils.typing import JSON



######################################################################
#
# File: b2sdk/v2/bucket.py
#
# Copyright 2021 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from __future__ import annotations

import typing

from b2sdk import v3
from b2sdk.v3.exception import BucketIdNotFound as v3BucketIdNotFound
from b2sdk.v2._compat import _file_infos_rename
from b2sdk._internal.http_constants import LIST_FILE_NAMES_MAX_LIMIT
from .exception import BucketIdNotFound
from .file_version import FileVersionFactory

if typing.TYPE_CHECKING:
    from b2sdk._internal.utils import Sha1HexDigest
    from b2sdk._internal.filter import Filter
    from .file_version import FileVersion


# Overridden to raise old style BucketIdNotFound exception
class Bucket(v3.Bucket):
    FILE_VERSION_FACTORY_CLASS = staticmethod(FileVersionFactory)

    def get_fresh_state(self) -> Bucket:
        try:
            return super().get_fresh_state()
        except v3BucketIdNotFound as e:
            raise BucketIdNotFound(e.bucket_id)

    @_file_infos_rename
    def upload_bytes(
        self,
        data_bytes,
        file_name,
        content_type=None,
        file_info: dict | None = None,
        progress_listener=None,
        encryption: v3.EncryptionSetting | None = None,
        file_retention: v3.FileRetentionSetting | None = None,
        legal_hold: v3.LegalHold | None = None,
        large_file_sha1: Sha1HexDigest | None = None,
        custom_upload_timestamp: int | None = None,
        cache_control: str | None = None,
        *args,
        **kwargs,
    ):
        return super().upload_bytes(
            data_bytes,
            file_name,
            content_type,
            file_info,
            progress_listener,
            encryption,
            file_retention,
            legal_hold,
            large_file_sha1,
            custom_upload_timestamp,
            cache_control,
            *args,
            **kwargs,
        )

    @_file_infos_rename
    def upload_local_file(
        self,
        local_file,
        file_name,
        content_type=None,
        file_info: dict | None = None,
        sha1_sum=None,
        min_part_size=None,
        progress_listener=None,
        encryption: v3.EncryptionSetting | None = None,
        file_retention: v3.FileRetentionSetting | None = None,
        legal_hold: v3.LegalHold | None = None,
        upload_mode: v3.UploadMode = v3.UploadMode.FULL,
        custom_upload_timestamp: int | None = None,
        cache_control: str | None = None,
        *args,
        **kwargs,
    ):
        return super().upload_local_file(
            local_file,
            file_name,
            content_type,
            file_info,
            sha1_sum,
            min_part_size,
            progress_listener,
            encryption,
            file_retention,
            legal_hold,
            upload_mode,
            custom_upload_timestamp,
            cache_control,
            *args,
            **kwargs,
        )

    def ls(
        self,
        folder_to_list: str = '',
        latest_only: bool = True,
        recursive: bool = False,
        fetch_count: int | None = LIST_FILE_NAMES_MAX_LIMIT,
        with_wildcard: bool = False,
        filters: typing.Sequence[Filter] = (),
        folder_to_list_can_be_a_file: bool = False,
        **kwargs,
    ) -> typing.Iterable[tuple[FileVersion, str]]:
        """
        Pretend that folders exist and yields the information about the files in a folder.

        B2 has a flat namespace for the files in a bucket, but there is a convention
        of using "/" as if there were folders.  This method searches through the
        flat namespace to find the files and "folders" that live within a given
        folder.

        When the `recursive` flag is set, lists all of the files in the given
        folder, and all of its sub-folders.

        :param folder_to_list: the name of the folder to list; must not start with "/".
                               Empty string means top-level folder
        :param latest_only: when ``False`` returns info about all versions of a file,
                            when ``True``, just returns info about the most recent versions
        :param recursive: if ``True``, list folders recursively
        :param fetch_count: how many entries to list per API call or ``None`` to use the default. Acceptable values: 1 - 10000
        :param with_wildcard: Accepts "*", "?", "[]" and "[!]" in folder_to_list, similarly to what shell does.
                              As of 1.19.0 it can only be enabled when recursive is also enabled.
                              Also, in this mode, folder_to_list is considered to be a filename or a pattern.
        :param filters: list of filters to apply to the files returned by the server.
        :param folder_to_list_can_be_a_file: if ``True``, folder_to_list can be a file, not just a folder
                                             This enabled default behavior of b2sdk.v3.Bucket.ls, in which for all
                                             paths that do not end with '/', first we try to check if file with this
                                             exact name exists, and only if it does not then we try to list files with
                                             this prefix.
        :rtype: generator[tuple[b2sdk.v2.FileVersion, str]]
        :returns: generator of (file_version, folder_name) tuples

        .. note::
            In case of `recursive=True`, folder_name is not returned.
        """
        if (
            not folder_to_list_can_be_a_file
            and folder_to_list
            and not folder_to_list.endswith('/')
            and not with_wildcard
        ):
            folder_to_list += '/'
        yield from super().ls(
            path=folder_to_list,
            latest_only=latest_only,
            recursive=recursive,
            fetch_count=fetch_count,
            with_wildcard=with_wildcard,
            filters=filters,
            **kwargs,
        )


# Overridden to use old style Bucket
class BucketFactory(v3.BucketFactory):
    BUCKET_CLASS = staticmethod(Bucket)



######################################################################
#
# File: b2sdk/v2/raw_simulator.py
#
# Copyright 2023 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from __future__ import annotations
import re
import time
from contextlib import suppress

from b2sdk import v3
from b2sdk.v2._compat import _file_infos_rename

from b2sdk.v2.exception import BadJson, NonExistentBucket, InvalidAuthToken, Unauthorized


class KeySimulator(v3.KeySimulator):
    """
    Hold information about one application key, which can be either
    a master application key, or one created with create_key().
    """

    def __init__(
        self,
        account_id,
        name,
        application_key_id,
        key,
        capabilities,
        expiration_timestamp_or_none,
        bucket_id_or_none,
        bucket_name_or_none,
        name_prefix_or_none,
    ):
        self.name = name
        self.account_id = account_id
        self.application_key_id = application_key_id
        self.key = key
        self.capabilities = capabilities
        self.expiration_timestamp_or_none = expiration_timestamp_or_none
        self.bucket_id_or_none = bucket_id_or_none
        self.bucket_name_or_none = bucket_name_or_none
        self.name_prefix_or_none = name_prefix_or_none

    def as_key(self):
        return dict(
            accountId=self.account_id,
            bucketId=self.bucket_id_or_none,
            applicationKeyId=self.application_key_id,
            capabilities=self.capabilities,
            expirationTimestamp=self.expiration_timestamp_or_none
            and self.expiration_timestamp_or_none * 1000,
            keyName=self.name,
            namePrefix=self.name_prefix_or_none,
        )

    def get_allowed(self):
        """
        Return the 'allowed' structure to include in the response from b2_authorize_account.
        """
        return dict(
            bucketId=self.bucket_id_or_none,
            bucketName=self.bucket_name_or_none,
            capabilities=self.capabilities,
            namePrefix=self.name_prefix_or_none,
        )


class BucketSimulator(v3.BucketSimulator):
    @_file_infos_rename
    def upload_file(
        self,
        upload_id: str,
        upload_auth_token: str,
        file_name: str,
        content_length: int,
        content_type: str,
        content_sha1: str,
        file_info: dict,
        data_stream,
        server_side_encryption: v3.EncryptionSetting | None = None,
        file_retention: v3.FileRetentionSetting | None = None,
        legal_hold: v3.LegalHold | None = None,
        custom_upload_timestamp: int | None = None,
        cache_control: str | None = None,
        *args,
        **kwargs,
    ):
        if cache_control is not None:
            file_info['b2-cache-control'] = cache_control
        return super().upload_file(
            upload_id,
            upload_auth_token,
            file_name,
            content_length,
            content_type,
            content_sha1,
            file_info,
            data_stream,
            server_side_encryption,
            file_retention,
            legal_hold,
            custom_upload_timestamp,
            *args,
            **kwargs,
        )


class RawSimulator(v3.RawSimulator):
    @classmethod
    @_file_infos_rename
    def get_upload_file_headers(
        cls,
        upload_auth_token: str,
        file_name: str,
        content_length: int,
        content_type: str,
        content_sha1: str,
        file_info: dict,
        server_side_encryption: v3.EncryptionSetting | None,
        file_retention: v3.FileRetentionSetting | None,
        legal_hold: v3.LegalHold | None,
        custom_upload_timestamp: int | None = None,
        cache_control: str | None = None,
        *args,
        **kwargs,
    ) -> dict:
        if cache_control is not None:
            file_info['b2-cache-control'] = cache_control
        return super().get_upload_file_headers(
            upload_auth_token,
            file_name,
            content_length,
            content_type,
            content_sha1,
            file_info,
            server_side_encryption,
            file_retention,
            legal_hold,
            custom_upload_timestamp,
            *args,
            **kwargs,
        )

    @_file_infos_rename
    def upload_file(
        self,
        upload_url: str,
        upload_auth_token: str,
        file_name: str,
        content_length: int,
        content_type: str,
        content_sha1: str,
        file_info: dict,
        data_stream,
        server_side_encryption: v3.EncryptionSetting | None = None,
        file_retention: v3.FileRetentionSetting | None = None,
        legal_hold: v3.LegalHold | None = None,
        custom_upload_timestamp: int | None = None,
        cache_control: str | None = None,
        *args,
        **kwargs,
    ):
        if cache_control is not None:
            file_info['b2-cache-control'] = cache_control
        return super().upload_file(
            upload_url,
            upload_auth_token,
            file_name,
            content_length,
            content_type,
            content_sha1,
            file_info,
            data_stream,
            server_side_encryption,
            file_retention,
            legal_hold,
            custom_upload_timestamp,
            *args,
            **kwargs,
        )

    def create_account(self):
        """
        Simulate creating an account.

        Return (accountId, masterApplicationKey) for a newly created account.
        """
        # Pick the IDs for the account and the key
        account_id = 'account-%d' % (self.account_counter,)
        master_key = 'masterKey-%d' % (self.account_counter,)
        self.account_counter += 1

        # Create the key
        self.key_id_to_key[account_id] = KeySimulator(
            account_id=account_id,
            name='master',
            application_key_id=account_id,
            key=master_key,
            capabilities=v3.ALL_CAPABILITIES,
            expiration_timestamp_or_none=None,
            bucket_id_or_none=None,
            bucket_name_or_none=None,
            name_prefix_or_none=None,
        )

        # Return the info
        return (account_id, master_key)

    def create_key(
        self,
        api_url,
        account_auth_token,
        account_id,
        capabilities,
        key_name,
        valid_duration_seconds,
        bucket_id,
        name_prefix,
    ):
        if not re.match(r'^[A-Za-z0-9-]{1,100}$', key_name):
            raise BadJson('illegal key name: ' + key_name)
        if valid_duration_seconds is not None:
            if valid_duration_seconds < 1 or valid_duration_seconds > self.MAX_DURATION_IN_SECONDS:
                raise BadJson(
                    'valid duration must be greater than 0, and less than 1000 days in seconds'
                )
        self._assert_account_auth(api_url, account_auth_token, account_id, 'writeKeys')

        if valid_duration_seconds is None:
            expiration_timestamp_or_none = None
        else:
            expiration_timestamp_or_none = int(time.time() + valid_duration_seconds)

        index = self.app_key_counter
        self.app_key_counter += 1
        application_key_id = 'appKeyId%d' % (index,)
        app_key = 'appKey%d' % (index,)
        bucket_name_or_none = None
        if bucket_id is not None:
            # It is possible for bucketId to be filled and bucketName to be empty.
            # It can happen when the bucket was deleted.
            with suppress(NonExistentBucket):
                bucket_name_or_none = self._get_bucket_by_id(bucket_id).bucket_name

        key_sim = KeySimulator(
            account_id=account_id,
            name=key_name,
            application_key_id=application_key_id,
            key=app_key,
            capabilities=capabilities,
            expiration_timestamp_or_none=expiration_timestamp_or_none,
            bucket_id_or_none=bucket_id,
            bucket_name_or_none=bucket_name_or_none,
            name_prefix_or_none=name_prefix,
        )
        self.key_id_to_key[application_key_id] = key_sim
        self.all_application_keys.append(key_sim)
        return key_sim.as_created_key()

    def _assert_account_auth(
        self, api_url, account_auth_token, account_id, capability, bucket_id=None, file_name=None
    ):
        key_sim = self.auth_token_to_key.get(account_auth_token)
        assert key_sim is not None
        assert api_url == self.API_URL
        assert account_id == key_sim.account_id
        if account_auth_token in self.expired_auth_tokens:
            raise InvalidAuthToken('auth token expired', 'auth_token_expired')
        if capability not in key_sim.capabilities:
            raise Unauthorized('', 'unauthorized')
        if key_sim.bucket_id_or_none is not None and key_sim.bucket_id_or_none != bucket_id:
            raise Unauthorized('', 'unauthorized')
        if key_sim.name_prefix_or_none is not None:
            if file_name is not None and not file_name.startswith(key_sim.name_prefix_or_none):
                raise Unauthorized('', 'unauthorized')

    def authorize_account(self, realm_url, application_key_id, application_key):
        key_sim = self.key_id_to_key.get(application_key_id)
        if key_sim is None:
            raise InvalidAuthToken('application key ID not valid', 'unauthorized')
        if application_key != key_sim.key:
            raise InvalidAuthToken('secret key is wrong', 'unauthorized')
        auth_token = 'auth_token_%d' % (self.auth_token_counter,)
        self.current_token = auth_token
        self.auth_token_counter += 1
        self.auth_token_to_key[auth_token] = key_sim
        allowed = key_sim.get_allowed()
        bucketId = allowed.get('bucketId')
        if (bucketId is not None) and (bucketId in self.bucket_id_to_bucket):
            allowed['bucketName'] = self.bucket_id_to_bucket[bucketId].bucket_name
        else:
            allowed['bucketName'] = None
        return dict(
            accountId=key_sim.account_id,
            authorizationToken=auth_token,
            apiInfo=dict(
                groupsApi=dict(),
                storageApi=dict(
                    apiUrl=self.API_URL,
                    downloadUrl=self.DOWNLOAD_URL,
                    recommendedPartSize=self.MIN_PART_SIZE,
                    absoluteMinimumPartSize=self.MIN_PART_SIZE,
                    allowed=allowed,
                    s3ApiUrl=self.S3_API_URL,
                    bucketId=allowed['bucketId'],
                    bucketName=allowed['bucketName'],
                    capabilities=allowed['capabilities'],
                    namePrefix=allowed['namePrefix'],
                ),
            ),
        )



######################################################################
#
# File: b2sdk/_internal/replication/setting.py
#
# Copyright 2022 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from __future__ import annotations

import re
from builtins import classmethod
from dataclasses import dataclass, field
from typing import ClassVar


@dataclass
class ReplicationRule:
    """
    Hold information about replication rule: destination bucket, priority,
    prefix and rule name.
    """

    DEFAULT_PRIORITY: ClassVar[int] = 128

    destination_bucket_id: str
    name: str
    file_name_prefix: str = ''
    is_enabled: bool = True
    priority: int = DEFAULT_PRIORITY
    include_existing_files: bool = False

    REPLICATION_RULE_REGEX: ClassVar = re.compile(r'^[a-zA-Z0-9_\-]{1,64}$')
    MIN_PRIORITY: ClassVar[int] = 1
    MAX_PRIORITY: ClassVar[int] = 2**31 - 1

    def __post_init__(self):
        if not self.destination_bucket_id:
            raise ValueError('destination_bucket_id is required')

        if not self.REPLICATION_RULE_REGEX.match(self.name):
            raise ValueError('replication rule name is invalid')

        if not (self.MIN_PRIORITY <= self.priority <= self.MAX_PRIORITY):
            raise ValueError(
                'priority should be within [%d, %d] interval'
                % (
                    self.MIN_PRIORITY,
                    self.MAX_PRIORITY,
                )
            )

    def as_dict(self) -> dict:
        return {
            'destinationBucketId': self.destination_bucket_id,
            'fileNamePrefix': self.file_name_prefix,
            'includeExistingFiles': self.include_existing_files,
            'isEnabled': self.is_enabled,
            'priority': self.priority,
            'replicationRuleName': self.name,
        }

    @classmethod
    def from_dict(cls, value_dict: dict) -> ReplicationRule:
        kwargs = {}
        for field_, protocolField in (
            ('destination_bucket_id', 'destinationBucketId'),
            ('name', 'replicationRuleName'),
            ('file_name_prefix', 'fileNamePrefix'),
            ('include_existing_files', 'includeExistingFiles'),
            ('is_enabled', 'isEnabled'),
            ('priority', 'priority'),
        ):
            value = value_dict.get(
                protocolField
            )  # refactor to := when dropping Python 3.7, maybe even dict expression
            if value is not None:
                kwargs[field_] = value
        return cls(**kwargs)


@dataclass
class ReplicationConfiguration:
    """
    Hold information about bucket replication configuration
    """

    # configuration as source:
    rules: list[ReplicationRule] = field(default_factory=list)
    source_key_id: str | None = None
    # configuration as destination:
    source_to_destination_key_mapping: dict[str, str] = field(default_factory=dict)

    def __post_init__(self):
        if self.rules and not self.source_key_id:
            raise ValueError('source_key_id must not be empty')

        for source, destination in self.source_to_destination_key_mapping.items():
            if not source or not destination:
                raise ValueError(
                    f'source_to_destination_key_mapping must not contain \
                     empty keys or values: ({source}, {destination})'
                )

    @property
    def is_source(self) -> bool:
        return bool(self.source_key_id)

    def get_source_configuration_as_dict(self) -> dict:
        return {
            'rules': self.rules,
            'source_key_id': self.source_key_id,
        }

    @property
    def is_destination(self) -> bool:
        return bool(self.source_to_destination_key_mapping)

    def get_destination_configuration_as_dict(self) -> dict:
        return {
            'source_to_destination_key_mapping': self.source_to_destination_key_mapping,
        }

    def as_dict(self) -> dict:
        """
        Represent the setting as a dict, for example:

        .. code-block:: python

            {
                "asReplicationSource": {
                    "replicationRules": [
                        {
                            "destinationBucketId": "c5f35d53a90a7ea284fb0719",
                            "fileNamePrefix": "",
                            "includeExistingFiles": True,
                            "isEnabled": true,
                            "priority": 1,
                            "replicationRuleName": "replication-us-west"
                        },
                        {
                            "destinationBucketId": "55f34d53a96a7ea284fb0719",
                            "fileNamePrefix": "",
                            "includeExistingFiles": True,
                            "isEnabled": true,
                            "priority": 2,
                            "replicationRuleName": "replication-us-west-2"
                        }
                    ],
                    "sourceApplicationKeyId": "10053d55ae26b790000000006"
                },
                "asReplicationDestination": {
                    "sourceToDestinationKeyMapping": {
                        "10053d55ae26b790000000045": "10053d55ae26b790000000004",
                        "10053d55ae26b790000000046": "10053d55ae26b790030000004"
                    }
                }
            }

        """

        result = {
            'asReplicationSource': {
                'replicationRules': [rule.as_dict() for rule in self.rules],
                'sourceApplicationKeyId': self.source_key_id,
            }
            if self.is_source
            else None,
            'asReplicationDestination': {
                'sourceToDestinationKeyMapping': self.source_to_destination_key_mapping,
            }
            if self.is_destination
            else None,
        }

        return result

    serialize_to_json_for_request = as_dict

    @classmethod
    def from_dict(cls, value_dict: dict) -> ReplicationConfiguration:
        source_dict = value_dict.get('asReplicationSource') or {}
        destination_dict = value_dict.get('asReplicationDestination') or {}

        return cls(
            rules=[
                ReplicationRule.from_dict(rule_dict)
                for rule_dict in source_dict.get('replicationRules', [])
            ],
            source_key_id=source_dict.get('sourceApplicationKeyId'),
            source_to_destination_key_mapping=destination_dict.get('sourceToDestinationKeyMapping')
            or {},
        )


@dataclass
class ReplicationConfigurationFactory:
    is_client_authorized_to_read: bool
    value: ReplicationConfiguration | None

    @classmethod
    def from_bucket_dict(cls, bucket_dict: dict) -> ReplicationConfigurationFactory:
        """
        Returns ReplicationConfigurationFactory for the given bucket dict
        retrieved from the api.
        """
        replication_dict = bucket_dict.get('replicationConfiguration') or {}
        value_dict = replication_dict.get('value') or {}

        return cls(
            is_client_authorized_to_read=replication_dict.get('isClientAuthorizedToRead', True),
            value=ReplicationConfiguration.from_dict(value_dict),
        )



######################################################################
#
# File: b2sdk/_internal/testing/helpers/buckets.py
#
# Copyright 2022 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from __future__ import annotations

import logging
import os
import platform
import random
import secrets
import time
from hashlib import sha256

from b2sdk._internal.http_constants import BUCKET_NAME_CHARS_UNIQ, BUCKET_NAME_LENGTH_RANGE

logger = logging.getLogger(__name__)

GENERAL_BUCKET_NAME_PREFIX = 'sdktst'
BUCKET_NAME_LENGTH = BUCKET_NAME_LENGTH_RANGE[1]
BUCKET_CREATED_AT_MILLIS = 'created_at_millis'
NODE_DESCRIPTION = f'{platform.node()}: {platform.platform()} {platform.python_version()}'


def get_seed() -> str:
    """
    Get seed for random number generator.

    The `WORKFLOW_ID` variable has to be set in the CI to uniquely identify
    the current workflow (including the attempt)
    """
    seed = ''.join(
        (
            os.getenv('WORKFLOW_ID', secrets.token_hex(8)),
            NODE_DESCRIPTION,
            str(time.time_ns()),
            os.getenv('PYTEST_XDIST_WORKER', 'gw0'),
        )
    )
    return sha256(seed.encode()).hexdigest()[:16]


RNG_SEED = get_seed()
RNG = random.Random(RNG_SEED)
RNG_COUNTER = 0


def random_token(length: int, chars: str = BUCKET_NAME_CHARS_UNIQ) -> str:
    return ''.join(RNG.choice(chars) for _ in range(length))


def bucket_name_part(length: int) -> str:
    assert length >= 1
    global RNG_COUNTER
    RNG_COUNTER += 1
    name_part = random_token(length, BUCKET_NAME_CHARS_UNIQ)
    logger.info('RNG_SEED: %s', RNG_SEED)
    logger.info('RNG_COUNTER: %i, length: %i', RNG_COUNTER, length)
    logger.info('name_part: %s', name_part)
    logger.info('WORKFLOW_ID: %s', os.getenv('WORKFLOW_ID'))
    return name_part


def get_bucket_name_prefix(rnd_len: int = 8, prefix: str = GENERAL_BUCKET_NAME_PREFIX) -> str:
    return prefix + bucket_name_part(rnd_len)



######################################################################
#
# File: b2sdk/_internal/utils/escape.py
#
# Copyright 2023 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################

from __future__ import annotations

import re
import shlex

# skip newline, tab
UNPRINTABLE_PATTERN = re.compile(r'[\x00-\x08\x0e-\x1f\x7f-\x9f]')


def unprintable_to_hex(s: str) -> str:
    """
    Replace unprintable chars in string with a hex representation.

    :param s: an arbitrary string, possibly with unprintable characters.
    :return: the string, with unprintable characters changed to hex (e.g., "\x07")
    """

    def hexify(match):
        return rf'\x{ord(match.group()):02x}'

    if s:
        return UNPRINTABLE_PATTERN.sub(hexify, s)
    return s


def escape_control_chars(s: str) -> str:
    """
    Replace unprintable chars in string with a hex representation AND shell quotes the string.

    :param s: an arbitrary string, possibly with unprintable characters.
    :return: the string, with unprintable characters changed to hex (e.g., "\x07")
    """
    if s:
        return shlex.quote(unprintable_to_hex(s))
    return s


def substitute_control_chars(s: str) -> tuple[str, bool]:
    """
    Replace unprintable chars in string with � unicode char

    :param s: an arbitrary string, possibly with unprintable characters.
    :return: tuple of the string with � replacements made and boolean indicated if chars were replaced
    """
    new_value = UNPRINTABLE_PATTERN.sub('�', s)
    return new_value, new_value != s



######################################################################
#
# File: b2sdk/_internal/filter.py
#
# Copyright 2024 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from __future__ import annotations

import fnmatch
from dataclasses import dataclass
from enum import Enum
from typing import Sequence


class FilterType(Enum):
    INCLUDE = 'include'
    EXCLUDE = 'exclude'


@dataclass
class Filter:
    type: FilterType
    pattern: str

    @classmethod
    def include(cls, pattern: str) -> Filter:
        return cls(type=FilterType.INCLUDE, pattern=pattern)

    @classmethod
    def exclude(cls, pattern: str) -> Filter:
        return cls(type=FilterType.EXCLUDE, pattern=pattern)


class FilterMatcher:
    """
    Holds a list of filters and matches a string (i.e. file name) against them.

    The order of filters matters. The *last* matching filter decides whether
    the string is included or excluded. If no filter matches, the string is
    included by default.

    If the given list of filters contains only INCLUDE filters, then it is
    assumed that all files are excluded by default. In this case, an additional
    EXCLUDE filter is prepended to the list.

    :param filters: list of filters
    """

    def __init__(self, filters: Sequence[Filter]):
        if filters and all(filter_.type == FilterType.INCLUDE for filter_ in filters):
            filters = [Filter(type=FilterType.EXCLUDE, pattern='*'), *filters]

        self.filters = filters

    def match(self, s: str) -> bool:
        include_file = True
        for filter_ in self.filters:
            matched = fnmatch.fnmatchcase(s, filter_.pattern)
            if matched and filter_.type == FilterType.INCLUDE:
                include_file = True
            elif matched and filter_.type == FilterType.EXCLUDE:
                include_file = False

        return include_file



######################################################################
#
# File: b2sdk/_internal/utils/thread_pool.py
#
# Copyright 2022 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from __future__ import annotations

import os
from concurrent.futures import Future, ThreadPoolExecutor
from typing import Callable

try:
    from typing_extensions import Protocol
except ImportError:
    from typing import Protocol

from b2sdk._internal.utils import B2TraceMetaAbstract


class DynamicThreadPoolExecutorProtocol(Protocol):
    def submit(self, fn: Callable, *args, **kwargs) -> Future: ...

    def set_size(self, max_workers: int) -> None:
        """Set the size of the thread pool."""

    def get_size(self) -> int:
        """Return the current size of the thread pool."""


class LazyThreadPool:
    """
    Lazily initialized thread pool.
    """

    _THREAD_POOL_FACTORY = ThreadPoolExecutor

    def __init__(self, max_workers: int | None = None, **kwargs):
        if max_workers is None:
            max_workers = min(
                32, (os.cpu_count() or 1) + 4
            )  # same default as in ThreadPoolExecutor
        self._max_workers = max_workers
        self._thread_pool: ThreadPoolExecutor | None = None
        super().__init__(**kwargs)

    def submit(self, fn: Callable, *args, **kwargs) -> Future:
        if self._thread_pool is None:
            self._thread_pool = self._THREAD_POOL_FACTORY(self._max_workers)
        return self._thread_pool.submit(fn, *args, **kwargs)

    def set_size(self, max_workers: int) -> None:
        """
        Set the size of the thread pool.

        This operation will block until all tasks in the current thread pool are completed.

        :param max_workers: New size of the thread pool
        :return: None
        """
        if self._max_workers == max_workers:
            return
        old_thread_pool = self._thread_pool
        self._thread_pool = self._THREAD_POOL_FACTORY(max_workers=max_workers)
        if old_thread_pool is not None:
            old_thread_pool.shutdown(wait=True)
        self._max_workers = max_workers

    def get_size(self) -> int:
        """Return the current size of the thread pool."""
        return self._max_workers


class ThreadPoolMixin(metaclass=B2TraceMetaAbstract):
    """
    Mixin class with ThreadPoolExecutor.
    """

    DEFAULT_THREAD_POOL_CLASS = LazyThreadPool

    def __init__(
        self,
        thread_pool: DynamicThreadPoolExecutorProtocol | None = None,
        max_workers: int | None = None,
        **kwargs,
    ):
        """
        :param thread_pool: thread pool to be used
        :param max_workers: maximum number of worker threads (ignored if thread_pool is not None)
        """
        self._thread_pool = (
            thread_pool
            if thread_pool is not None
            else self.DEFAULT_THREAD_POOL_CLASS(max_workers=max_workers)
        )
        self._max_workers = max_workers
        super().__init__(**kwargs)

    def set_thread_pool_size(self, max_workers: int) -> None:
        """
        Set the size of the thread pool.

        This operation will block until all tasks in the current thread pool are completed.

        :param max_workers: New size of the thread pool
        :return: None
        """
        return self._thread_pool.set_size(max_workers)

    def get_thread_pool_size(self) -> int:
        return self._thread_pool.get_size()



######################################################################
#
# File: b2sdk/_internal/http_constants.py
#
# Copyright 2021 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from __future__ import annotations

import string

# These constants are needed in different modules, so they are stored in this module, that
# imports nothing, thus avoiding circular imports

# https://www.backblaze.com/docs/cloud-storage-buckets#bucket-names
BUCKET_NAME_CHARS = string.ascii_lowercase + string.digits + '-'
BUCKET_NAME_CHARS_UNIQ = string.ascii_lowercase + string.digits + '-'
BUCKET_NAME_LENGTH_RANGE = (6, 63)

LIST_FILE_NAMES_MAX_LIMIT = 10000  # https://www.backblaze.com/b2/docs/b2_list_file_names.html

FILE_INFO_HEADER_PREFIX = 'X-Bz-Info-'
FILE_INFO_HEADER_PREFIX_LOWER = FILE_INFO_HEADER_PREFIX.lower()

# Standard names for file info entries
SRC_LAST_MODIFIED_MILLIS = 'src_last_modified_millis'

# SHA-1 hash key for large files
LARGE_FILE_SHA1 = 'large_file_sha1'

# Special X-Bz-Content-Sha1 value to verify checksum at the end
HEX_DIGITS_AT_END = 'hex_digits_at_end'

# Identifying SSE_C keys
SSE_C_KEY_ID_FILE_INFO_KEY_NAME = 'sse_c_key_id'
SSE_C_KEY_ID_HEADER = FILE_INFO_HEADER_PREFIX + SSE_C_KEY_ID_FILE_INFO_KEY_NAME

# Default part sizes
MEGABYTE = 1000 * 1000
GIGABYTE = 1000 * MEGABYTE
DEFAULT_MIN_PART_SIZE = 5 * MEGABYTE
DEFAULT_RECOMMENDED_UPLOAD_PART_SIZE = 100 * MEGABYTE
DEFAULT_MAX_PART_SIZE = 5 * GIGABYTE



######################################################################
#
# File: b2sdk/_internal/__init__.py
#
# Copyright 2023 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
"""
b2sdk._internal package contains internal modules, and should not be used directly.

Please use chosen apiver package instead, e.g. b2sdk.v3
"""



######################################################################
#
# File: b2sdk/v2/api_config.py
#
# Copyright 2025 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from b2sdk import v3
from .raw_api import B2RawHTTPApi


class B2HttpApiConfig(v3.B2HttpApiConfig):
    DEFAULT_RAW_API_CLASS = B2RawHTTPApi


DEFAULT_HTTP_API_CONFIG = B2HttpApiConfig()



######################################################################
#
# File: b2sdk/v2/account_info.py
#
# Copyright 2023 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from __future__ import annotations

import json

from b2sdk import v3
from .exception import MissingAccountData


class _OldAllowedMixin:
    DEFAULT_ALLOWED = dict(
        bucketId=None,
        bucketName=None,
        capabilities=v3.ALL_CAPABILITIES,
        namePrefix=None,
    )

    @classmethod
    def allowed_is_valid(cls, allowed):
        return (
            ('bucketId' in allowed)
            and ('bucketName' in allowed)
            and ((allowed['bucketId'] is not None) or (allowed['bucketName'] is None))
            and ('capabilities' in allowed)
            and ('namePrefix' in allowed)
        )


class AbstractAccountInfo(_OldAllowedMixin, v3.AbstractAccountInfo):
    def list_bucket_names_ids(self):
        return []  # Removed @abstractmethod decorator


class UrlPoolAccountInfo(_OldAllowedMixin, v3.UrlPoolAccountInfo):
    pass


class InMemoryAccountInfo(_OldAllowedMixin, v3.InMemoryAccountInfo):
    pass


class SqliteAccountInfo(_OldAllowedMixin, v3.SqliteAccountInfo):
    def get_allowed(self):
        """
        Return 'allowed' dictionary info.
        Example:

        .. code-block:: python

            {
                "bucketId": null,
                "bucketName": null,
                "capabilities": [
                    "listKeys",
                    "writeKeys"
                ],
                "namePrefix": null
            }

        The 'allowed' column was not in the original schema, so it may be NULL.

        :rtype: dict
        """
        allowed_json = self._get_account_info_or_raise('allowed')
        if allowed_json is None:
            return self.DEFAULT_ALLOWED

        allowed = json.loads(allowed_json)

        # convert a multi-bucket key to a single bucket

        if 'buckets' in allowed:
            buckets = allowed.pop('buckets')
            if buckets and len(buckets) > 1:
                raise MissingAccountData(
                    'Multi-bucket keys cannot be used with the current sdk version'
                )

            allowed['bucketId'] = buckets[0]['id'] if buckets else None
            allowed['bucketName'] = buckets[0]['name'] if buckets else None

        return allowed


class StubAccountInfo(_OldAllowedMixin, v3.StubAccountInfo):
    pass



######################################################################
#
# File: b2sdk/_internal/testing/fixtures/buckets.py
#
# Copyright 2021 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from __future__ import annotations

import secrets

import pytest

from b2sdk._internal.testing.helpers.bucket_manager import BucketManager
from b2sdk._internal.testing.helpers.buckets import (
    GENERAL_BUCKET_NAME_PREFIX,
    get_bucket_name_prefix,
)


def pytest_addoption(parser):
    """Add a flag for not cleaning up old buckets"""
    parser.addoption(
        '--dont-cleanup-old-buckets',
        action='store_true',
        default=False,
    )


@pytest.fixture(scope='session')
def dont_cleanup_old_buckets(request):
    return request.config.getoption('--dont-cleanup-old-buckets')


@pytest.fixture(scope='session')
def general_bucket_name_prefix():
    return GENERAL_BUCKET_NAME_PREFIX


@pytest.fixture(scope='session')
def bucket_name_prefix(general_bucket_name_prefix):
    return get_bucket_name_prefix(8, general_bucket_name_prefix)


@pytest.fixture(scope='session')
def bucket_manager(
    bucket_name_prefix, general_bucket_name_prefix, dont_cleanup_old_buckets, _b2_api
):
    manager = BucketManager(
        dont_cleanup_old_buckets,
        _b2_api,
        current_run_prefix=bucket_name_prefix,
        general_prefix=general_bucket_name_prefix,
    )
    yield manager
    manager.clean_buckets()


@pytest.fixture
def bucket(bucket_name_prefix, bucket_manager):
    bucket = bucket_manager.create_bucket()
    yield bucket
    bucket_manager.clean_bucket(bucket)


@pytest.fixture
def b2_subfolder(bucket, request):
    subfolder_name = f'{request.node.name}_{secrets.token_urlsafe(4)}'
    return f'b2://{bucket.name}/{subfolder_name}'



######################################################################
#
# File: b2sdk/v0/sync.py
#
# Copyright 2019 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from __future__ import annotations

import logging

from b2sdk.v1.exception import CommandError
from b2sdk.v1.exception import DestFileNewer as DestFileNewerV1
from b2sdk.v1 import trace_call
from .exception import DestFileNewer
from b2sdk.v1.exception import InvalidArgument, IncompleteSync
from b2sdk.v1 import NewerFileSyncMode, CompareVersionMode
from b2sdk.v1 import KeepOrDeleteMode
from b2sdk.v1 import DEFAULT_SCAN_MANAGER
from b2sdk.v1 import SyncReport
from b2sdk.v1 import Synchronizer as SynchronizerV1
from b2sdk.v1 import (
    AbstractSyncEncryptionSettingsProvider,
    SERVER_DEFAULT_SYNC_ENCRYPTION_SETTINGS_PROVIDER,
)

logger = logging.getLogger(__name__)


class Synchronizer(SynchronizerV1):
    """
    This is wrapper for newer version and will return the v0 style exceptions
    """

    def __init__(self, *args, **kwargs):
        try:
            super().__init__(*args, **kwargs)
        except InvalidArgument as e:
            raise CommandError(f'--{e.parameter_name} {e.message}')

    def _make_file_sync_actions(self, *args, **kwargs):
        try:
            yield from super()._make_file_sync_actions(*args, **kwargs)
        except DestFileNewerV1 as e:
            raise DestFileNewer(e.dest_file, e.source_file, e.dest_prefix, e.source_prefix)

    def sync_folders(self, *args, **kwargs):
        try:
            super().sync_folders(*args, **kwargs)
        except InvalidArgument as e:
            raise CommandError(f'--{e.parameter_name} {e.message}')
        except IncompleteSync as e:
            raise CommandError(str(e))


def get_synchronizer_from_args(
    args,
    max_workers,
    policies_manager=DEFAULT_SCAN_MANAGER,
    dry_run=False,
    allow_empty_source=False,
):
    if args.replaceNewer and args.skipNewer:
        raise CommandError('--skipNewer and --replaceNewer are incompatible')
    elif args.replaceNewer:
        newer_file_mode = NewerFileSyncMode.REPLACE
    elif args.skipNewer:
        newer_file_mode = NewerFileSyncMode.SKIP
    else:
        newer_file_mode = NewerFileSyncMode.RAISE_ERROR

    if args.delete and (args.keepDays is not None):
        raise CommandError('--delete and --keepDays are incompatible')

    if args.compareVersions == 'none':
        compare_version_mode = CompareVersionMode.NONE
    elif args.compareVersions == 'modTime':
        compare_version_mode = CompareVersionMode.MODTIME
    elif args.compareVersions == 'size':
        compare_version_mode = CompareVersionMode.SIZE
    elif args.compareVersions is None:
        compare_version_mode = CompareVersionMode.MODTIME
    else:
        raise CommandError('Invalid option for --compareVersions')
    compare_threshold = args.compareThreshold

    keep_days = None

    if args.delete:
        keep_days_or_delete = KeepOrDeleteMode.DELETE
    elif args.keepDays:
        keep_days_or_delete = KeepOrDeleteMode.KEEP_BEFORE_DELETE
        keep_days = args.keepDays
    else:
        keep_days_or_delete = KeepOrDeleteMode.NO_DELETE

    return Synchronizer(
        max_workers,
        policies_manager=policies_manager,
        dry_run=dry_run,
        allow_empty_source=allow_empty_source,
        newer_file_mode=newer_file_mode,
        keep_days_or_delete=keep_days_or_delete,
        compare_version_mode=compare_version_mode,
        compare_threshold=compare_threshold,
        keep_days=keep_days,
    )


def make_folder_sync_actions(
    source_folder,
    dest_folder,
    args,
    now_millis,
    reporter,
    policies_manager=DEFAULT_SCAN_MANAGER,
    encryption_settings_provider: AbstractSyncEncryptionSettingsProvider = SERVER_DEFAULT_SYNC_ENCRYPTION_SETTINGS_PROVIDER,
):
    """
    This is deprecated. Use the new Synchronizer class.
    Yields a sequence of actions that will sync the destination
    folder to the source folder.

    :param source_folder: source folder object
    :type source_folder: b2sdk._internal.scan.folder.AbstractFolder
    :param dest_folder: destination folder object
    :type dest_folder: b2sdk._internal.scan.folder.AbstractFolder
    :param args: an object which holds command line arguments
    :param now_millis: current time in milliseconds
    :type now_millis: int
    :param reporter: reporter object
    :param policies_manager: policies manager object
    :param encryption_settings_provider: encryption settings provider
    :type encryption_settings_provider: AbstractSyncEncryptionSettingsProvider
    """
    synchronizer = get_synchronizer_from_args(
        args,
        1,
        policies_manager=policies_manager,
        dry_run=False,
        allow_empty_source=False,
    )
    try:
        return synchronizer.make_folder_sync_actions(
            source_folder,
            dest_folder,
            now_millis,
            reporter,
            policies_manager=policies_manager,
            encryption_settings_provider=encryption_settings_provider,
        )
    except InvalidArgument as e:
        raise CommandError(f'--{e.parameter_name} {e.message}')


@trace_call(logger)
def sync_folders(
    source_folder,
    dest_folder,
    args,
    now_millis,
    stdout,
    no_progress,
    max_workers,
    policies_manager=DEFAULT_SCAN_MANAGER,
    dry_run=False,
    allow_empty_source=False,
    encryption_settings_provider: AbstractSyncEncryptionSettingsProvider = SERVER_DEFAULT_SYNC_ENCRYPTION_SETTINGS_PROVIDER,
):
    """
    This is deprecated. Use the new Synchronizer class.

    source is also in the destination.  Deletes any file versions
    in the destination older than history_days.

    :param source_folder: source folder object
    :type source_folder: b2sdk._internal.scan.folder.AbstractFolder
    :param dest_folder: destination folder object
    :type dest_folder: b2sdk._internal.scan.folder.AbstractFolder
    :param args: an object which holds command line arguments
    :param now_millis: current time in milliseconds
    :type now_millis: int
    :param stdout: standard output file object
    :param no_progress: if True, do not show progress
    :type no_progress: bool
    :param max_workers: max number of workers
    :type max_workers: int
    :param policies_manager: policies manager object
    :param dry_run:
    :type dry_run: bool
    :param allow_empty_source: if True, do not check whether source folder is empty
    :type allow_empty_source: bool
    :param encryption_settings_provider: encryption settings provider
    :type encryption_settings_provider: AbstractSyncEncryptionSettingsProvider
    """
    synchronizer = get_synchronizer_from_args(
        args,
        max_workers,
        policies_manager=policies_manager,
        dry_run=dry_run,
        allow_empty_source=allow_empty_source,
    )
    with SyncReport(stdout, no_progress) as reporter:
        synchronizer.sync_folders(
            source_folder,
            dest_folder,
            now_millis,
            reporter,
            encryption_settings_provider=encryption_settings_provider,
        )



######################################################################
#
# File: b2sdk/_internal/transfer/emerge/executor.py
#
# Copyright 2020 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from __future__ import annotations

import logging
import threading
from abc import ABCMeta, abstractmethod
from typing import TYPE_CHECKING

from b2sdk._internal.encryption.setting import EncryptionSetting
from b2sdk._internal.exception import MaxFileSizeExceeded
from b2sdk._internal.file_lock import NO_RETENTION_FILE_SETTING, FileRetentionSetting, LegalHold
from b2sdk._internal.http_constants import LARGE_FILE_SHA1
from b2sdk._internal.transfer.outbound.large_file_upload_state import LargeFileUploadState
from b2sdk._internal.transfer.outbound.upload_source import UploadSourceStream

AUTO_CONTENT_TYPE = 'b2/x-auto'

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from b2sdk._internal.transfer.emerge.planner.part_definition import UploadEmergePartDefinition
    from b2sdk._internal.transfer.emerge.planner.planner import StreamingEmergePlan


class EmergeExecutor:
    def __init__(self, services):
        self.services = services

    def execute_emerge_plan(
        self,
        emerge_plan,
        bucket_id,
        file_name,
        content_type,
        file_info,
        progress_listener,
        continue_large_file_id=None,
        max_queue_size=None,
        encryption: EncryptionSetting | None = None,
        file_retention: FileRetentionSetting | None = None,
        legal_hold: LegalHold | None = None,
        custom_upload_timestamp: int | None = None,
    ):
        if emerge_plan.is_large_file():
            execution = LargeFileEmergeExecution(
                self.services,
                bucket_id,
                file_name,
                content_type,
                file_info,
                progress_listener,
                encryption=encryption,
                file_retention=file_retention,
                legal_hold=legal_hold,
                continue_large_file_id=continue_large_file_id,
                max_queue_size=max_queue_size,
                custom_upload_timestamp=custom_upload_timestamp,
            )
        else:
            if continue_large_file_id is not None:
                raise ValueError('Cannot resume emerging single part plan.')
            execution = SmallFileEmergeExecution(
                self.services,
                bucket_id,
                file_name,
                content_type,
                file_info,
                progress_listener,
                encryption=encryption,
                file_retention=file_retention,
                legal_hold=legal_hold,
                custom_upload_timestamp=custom_upload_timestamp,
            )
        return execution.execute_plan(emerge_plan)


class BaseEmergeExecution(metaclass=ABCMeta):
    DEFAULT_CONTENT_TYPE = AUTO_CONTENT_TYPE

    def __init__(
        self,
        services,
        bucket_id,
        file_name,
        content_type,
        file_info,
        progress_listener,
        encryption: EncryptionSetting | None = None,
        file_retention: FileRetentionSetting | None = None,
        legal_hold: LegalHold | None = None,
        custom_upload_timestamp: int | None = None,
    ):
        self.services = services
        self.bucket_id = bucket_id
        self.file_name = file_name
        self.content_type = content_type
        self.file_info = file_info
        self.progress_listener = progress_listener
        self.encryption = encryption
        self.file_retention = file_retention
        self.legal_hold = legal_hold
        self.custom_upload_timestamp = custom_upload_timestamp

    @abstractmethod
    def execute_plan(self, emerge_plan):
        pass


class SmallFileEmergeExecution(BaseEmergeExecution):
    def execute_plan(self, emerge_plan):
        emerge_parts = list(emerge_plan.emerge_parts)
        assert len(emerge_parts) == 1
        emerge_part = emerge_parts[0]
        execution_step_factory = SmallFileEmergeExecutionStepFactory(self, emerge_part)
        execution_step = execution_step_factory.get_execution_step()
        future = execution_step.execute()
        return future.result()


class LargeFileEmergeExecution(BaseEmergeExecution):
    MAX_LARGE_FILE_SIZE = 10 * 1000 * 1000 * 1000 * 1000  # 10 TB

    def __init__(
        self,
        services,
        bucket_id,
        file_name,
        content_type,
        file_info,
        progress_listener,
        encryption: EncryptionSetting | None = None,
        file_retention: FileRetentionSetting | None = None,
        legal_hold: LegalHold | None = None,
        continue_large_file_id=None,
        max_queue_size=None,
        custom_upload_timestamp: int | None = None,
    ):
        super().__init__(
            services,
            bucket_id,
            file_name,
            content_type,
            file_info,
            progress_listener,
            encryption=encryption,
            file_retention=file_retention,
            legal_hold=legal_hold,
            custom_upload_timestamp=custom_upload_timestamp,
        )
        self.continue_large_file_id = continue_large_file_id
        self.max_queue_size = max_queue_size
        self._semaphore = None
        if self.max_queue_size is not None:
            self._semaphore = threading.Semaphore(self.max_queue_size)

    def execute_plan(self, emerge_plan: StreamingEmergePlan):
        total_length = emerge_plan.get_total_length()
        encryption = self.encryption

        if total_length is not None and total_length > self.MAX_LARGE_FILE_SIZE:
            raise MaxFileSizeExceeded(total_length, self.MAX_LARGE_FILE_SIZE)

        plan_id = emerge_plan.get_plan_id()

        file_info = dict(self.file_info) if self.file_info is not None else {}
        if plan_id is not None:
            file_info['plan_id'] = plan_id

        self.progress_listener.set_total_bytes(total_length or 0)

        emerge_parts_dict = None
        if total_length is not None:
            emerge_parts_dict = dict(emerge_plan.enumerate_emerge_parts())

        unfinished_file, finished_parts = self._get_unfinished_file_and_parts(
            self.bucket_id,
            self.file_name,
            file_info,
            self.continue_large_file_id,
            encryption=encryption,
            file_retention=self.file_retention,
            legal_hold=self.legal_hold,
            emerge_parts_dict=emerge_parts_dict,
            custom_upload_timestamp=self.custom_upload_timestamp,
        )

        if unfinished_file is None:
            if self.content_type is None:
                content_type = self.DEFAULT_CONTENT_TYPE
            else:
                content_type = self.content_type
            unfinished_file = self.services.large_file.start_large_file(
                self.bucket_id,
                self.file_name,
                content_type,
                file_info,
                encryption=encryption,
                file_retention=self.file_retention,
                legal_hold=self.legal_hold,
            )
        file_id = unfinished_file.file_id

        large_file_upload_state = LargeFileUploadState(self.progress_listener)

        part_futures = []
        for part_number, emerge_part in emerge_plan.enumerate_emerge_parts():
            execution_step_factory = LargeFileEmergeExecutionStepFactory(
                self,
                emerge_part,
                part_number,
                file_id,
                large_file_upload_state,
                finished_parts=finished_parts,
                # it already knows encryption from BaseMergeExecution being passed as self
            )
            execution_step = execution_step_factory.get_execution_step()
            future = self._execute_step(execution_step)
            part_futures.append(future)

        # Collect the sha1 checksums of the parts as the uploads finish.
        # If any of them raised an exception, that same exception will
        # be raised here by result()
        part_sha1_array = [f.result()['contentSha1'] for f in part_futures]

        # Finish the large file
        response = self.services.session.finish_large_file(file_id, part_sha1_array)
        return self.services.api.file_version_factory.from_api_response(response)

    def _execute_step(self, execution_step: UploadPartExecutionStep):
        semaphore = self._semaphore
        if semaphore is None:
            return execution_step.execute()
        else:
            semaphore.acquire()
            try:
                future = execution_step.execute()
            except:
                semaphore.release()
                raise
            else:
                future.add_done_callback(lambda f: semaphore.release())
            return future

    def _get_unfinished_file_and_parts(
        self,
        bucket_id,
        file_name,
        file_info,
        continue_large_file_id,
        encryption: EncryptionSetting,
        file_retention: FileRetentionSetting | None = None,
        legal_hold: LegalHold | None = None,
        emerge_parts_dict=None,
        custom_upload_timestamp: int | None = None,
    ):
        if 'listFiles' not in self.services.session.account_info.get_allowed()['capabilities']:
            return None, {}

        unfinished_file = None
        finished_parts = {}

        if continue_large_file_id is not None:
            unfinished_file = self.services.large_file.get_unfinished_large_file(
                bucket_id,
                continue_large_file_id,
                prefix=file_name,
            )
            if unfinished_file.file_info != file_info:
                raise ValueError(
                    'Cannot manually resume unfinished large file with different file_info'
                )
            finished_parts = {
                part.part_number: part
                for part in self.services.large_file.list_parts(continue_large_file_id)
            }
        elif 'plan_id' in file_info:
            assert emerge_parts_dict is not None
            unfinished_file, finished_parts = self._find_unfinished_file_by_plan_id(
                bucket_id,
                file_name,
                file_info,
                emerge_parts_dict,
                encryption,
                file_retention,
                legal_hold,
                custom_upload_timestamp=custom_upload_timestamp,
            )
        elif emerge_parts_dict is not None:
            unfinished_file, finished_parts = self._match_unfinished_file_if_possible(
                bucket_id,
                file_name,
                file_info,
                emerge_parts_dict,
                encryption,
                file_retention,
                legal_hold,
                custom_upload_timestamp=custom_upload_timestamp,
            )
        return unfinished_file, finished_parts

    def _find_matching_unfinished_file(
        self,
        bucket_id,
        file_name,
        file_info,
        emerge_parts_dict,
        encryption: EncryptionSetting,
        file_retention: FileRetentionSetting | None = None,
        legal_hold: LegalHold | None = None,
        custom_upload_timestamp: int | None = None,
        check_file_info_without_large_file_sha1: bool | None = False,
        eager_mode: bool | None = False,
    ):
        """
        Search for a matching unfinished large file in the specified bucket.

        In case a matching file is found but has inconsistencies (for example, mismatching file info or encryption settings),
        mismatches are logged.

        :param bucket_id: The identifier of the bucket where the unfinished file resides.
        :param file_name: The name of the file to be matched.
        :param file_info: Information about the file to be uploaded.
        :param emerge_parts_dict: A dictionary containing the parts of the file to be emerged.
        :param encryption: The encryption settings for the file.
        :param file_retention: The retention settings for the file, if any.
        :param legal_hold: The legal hold status of the file, if any.
        :param custom_upload_timestamp: The custom timestamp for the upload, if any.
        :param check_file_info_without_large_file_sha1: A flag indicating whether the file information should be checked without the `large_file_sha1`.
        :param eager_mode: A flag indicating whether the first matching file should be returned.

        :return: A tuple of the best matching unfinished file and its finished parts. If no match is found, returns `None`.
        """

        file_retention = file_retention or NO_RETENTION_FILE_SETTING
        best_match_file = None
        best_match_parts = {}
        best_match_parts_len = 0

        for file_ in self.services.large_file.list_unfinished_large_files(
            bucket_id, prefix=file_name
        ):
            if file_.file_name != file_name:
                logger.debug('Rejecting %s: file name mismatch', file_.file_id)
                continue

            if file_.file_info != file_info:
                if check_file_info_without_large_file_sha1:
                    file_info_without_large_file_sha1 = self._get_file_info_without_large_file_sha1(
                        file_info
                    )
                    if (
                        file_info_without_large_file_sha1
                        != self._get_file_info_without_large_file_sha1(file_.file_info)
                    ):
                        logger.debug(
                            'Rejecting %s: file info mismatch after dropping `large_file_sha1`',
                            file_.file_id,
                        )
                        continue
                else:
                    logger.debug('Rejecting %s: file info mismatch', file_.file_id)
                    continue

            if encryption is not None and encryption != file_.encryption:
                logger.debug('Rejecting %s: encryption mismatch', file_.file_id)
                continue

            if legal_hold is None:
                if LegalHold.UNSET != file_.legal_hold:
                    logger.debug('Rejecting %s: legal hold mismatch (not unset)', file_.file_id)
                    continue
            elif legal_hold != file_.legal_hold:
                logger.debug('Rejecting %s: legal hold mismatch', file_.file_id)
                continue

            if file_retention != file_.file_retention:
                logger.debug('Rejecting %s: retention mismatch', file_.file_id)
                continue

            if (
                custom_upload_timestamp is not None
                and file_.upload_timestamp != custom_upload_timestamp
            ):
                logger.debug('Rejecting %s: custom_upload_timestamp mismatch', file_.file_id)
                continue

            finished_parts = {}
            conflict_detected = False

            for part in self.services.large_file.list_parts(file_.file_id):
                emerge_part = emerge_parts_dict.get(part.part_number)

                if emerge_part is None:
                    # something is wrong - we have a part that we don't know about
                    # so we can't resume this upload
                    logger.debug(
                        'Rejecting %s: part %s not found in emerge parts, giving up.',
                        file_.file_id,
                        part.part_number,
                    )
                    conflict_detected = True
                    break

                # Compare part sizes
                if emerge_part.get_length() != part.content_length:
                    logger.debug(
                        'Rejecting %s: part %s size mismatch', file_.file_id, part.part_number
                    )
                    conflict_detected = True
                    break  # part size doesn't match - so we reupload

                # Compare part hashes
                if emerge_part.is_hashable() and emerge_part.get_sha1() != part.content_sha1:
                    logger.debug(
                        'Rejecting %s: part %s sha1 mismatch', file_.file_id, part.part_number
                    )
                    conflict_detected = True
                    break  # part.sha1 doesn't match - so we reupload

                finished_parts[part.part_number] = part

            if conflict_detected:
                continue

            finished_parts_len = len(finished_parts)

            if best_match_file is None or finished_parts_len > best_match_parts_len:
                best_match_file = file_
                best_match_parts = finished_parts
                best_match_parts_len = finished_parts_len

            if eager_mode and best_match_file is not None:
                break

        return best_match_file, best_match_parts

    def _find_unfinished_file_by_plan_id(
        self,
        bucket_id,
        file_name,
        file_info,
        emerge_parts_dict,
        encryption: EncryptionSetting,
        file_retention: FileRetentionSetting | None = None,
        legal_hold: LegalHold | None = None,
        custom_upload_timestamp: int | None = None,
    ):
        """
        Search for a matching unfinished large file by plan_id in the specified bucket.

        This function aims to locate a matching unfinished large file using the plan_id and the supplied parameters.
        It's used to resume an interrupted upload, centralizing the shared logic between `_find_unfinished_file_by_plan_id`
        and `_match_unfinished_file_if_possible`.

        In case a matching file is found but has inconsistencies (for example, mismatching file info or encryption settings),
        the function checks if 'plan_id' is in file_info, as this is a prerequisite.

        :param bucket_id: The identifier of the bucket where the unfinished file resides.
        :param file_name: The name of the file to be matched.
        :param file_info: Information about the file to be uploaded.
        :param emerge_parts_dict: A dictionary containing the parts of the file to be emerged.
        :param encryption: The encryption settings for the file.
        :param file_retention: The retention settings for the file, if any.
        :param legal_hold: The legal hold status of the file, if any.
        :param custom_upload_timestamp: The custom timestamp for the upload, if any.

        :return: A tuple of the best matching unfinished file and its finished parts. If no match is found, it returns `None`.
        """
        if 'plan_id' not in file_info:
            raise ValueError("The 'plan_id' key must be in file_info dictionary.")

        return self._find_matching_unfinished_file(
            bucket_id=bucket_id,
            file_name=file_name,
            file_info=file_info,
            emerge_parts_dict=emerge_parts_dict,
            encryption=encryption,
            file_retention=file_retention or NO_RETENTION_FILE_SETTING,
            legal_hold=legal_hold,
            custom_upload_timestamp=custom_upload_timestamp,
            check_file_info_without_large_file_sha1=False,
        )

    @classmethod
    def _get_file_info_without_large_file_sha1(
        cls,
        file_info: dict[str, str] | None,
    ) -> dict[str, str] | None:
        if not file_info or LARGE_FILE_SHA1 not in file_info:
            return file_info
        out_file_info = dict(file_info)
        del out_file_info[LARGE_FILE_SHA1]
        return out_file_info

    def _match_unfinished_file_if_possible(
        self,
        bucket_id,
        file_name,
        file_info,
        emerge_parts_dict,
        encryption: EncryptionSetting,
        file_retention: FileRetentionSetting | None = None,
        legal_hold: LegalHold | None = None,
        custom_upload_timestamp: int | None = None,
    ):
        """
        Scan for a suitable unfinished large file in the specified bucket to resume upload.

        This function examines each unfinished large file for a possible match with the provided
        parameters. This enables resumption of an interrupted upload by reusing the unfinished file,
        provided that file's info and additional parameters match.

        Along with the filename and file info, additional parameters like encryption, file retention,
        legal hold, custom upload timestamp, and cache control are compared for a match. The
        'emerge_parts_dict' is also cross-checked for matching file parts.

        Function is eager to find a match, and will return the first match it finds. If no match is
        found, it returns `None`.

        :param bucket_id: The identifier of the bucket containing the unfinished file.
        :param file_name: The name of the file to find.
        :param file_info: Information about the file to be uploaded.
        :param emerge_parts_dict: A dictionary of the parts of the file to be emerged.
        :param encryption: The encryption settings for the file.
        :param file_retention: The retention settings for the file, if applicable.
        :param legal_hold: The legal hold status of the file, if applicable.
        :param custom_upload_timestamp: The custom timestamp for the upload, if set.

        :return: A tuple of the best matching unfinished file and its finished parts. If no match is found, returns `None`.
        """
        logger.debug('Checking for matching unfinished large files for %s...', file_name)

        file_, finished_parts = self._find_matching_unfinished_file(
            bucket_id,
            file_name,
            file_info,
            emerge_parts_dict,
            encryption,
            file_retention,
            legal_hold,
            custom_upload_timestamp,
            check_file_info_without_large_file_sha1=True,
            eager_mode=True,
        )

        if file_ is None:
            logger.debug('No matching unfinished files found.')
            return None, {}

        logger.debug(
            'Unfinished file %s matches with %i finished parts', file_.file_id, len(finished_parts)
        )
        return file_, finished_parts


class BaseExecutionStepFactory(metaclass=ABCMeta):
    def __init__(self, emerge_execution, emerge_part):
        self.emerge_execution = emerge_execution
        self.emerge_part = emerge_part

    @abstractmethod
    def create_copy_execution_step(self, copy_range):
        pass

    @abstractmethod
    def create_upload_execution_step(self, stream_opener, stream_length=None, stream_sha1=None):
        pass

    def get_execution_step(self):
        return self.emerge_part.get_execution_step(self)


class SmallFileEmergeExecutionStepFactory(BaseExecutionStepFactory):
    def create_copy_execution_step(self, copy_range):
        return CopyFileExecutionStep(self.emerge_execution, copy_range)

    def create_upload_execution_step(self, stream_opener, stream_length=None, stream_sha1=None):
        return UploadFileExecutionStep(
            self.emerge_execution,
            stream_opener,
            stream_length=stream_length,
            stream_sha1=stream_sha1,
        )


class LargeFileEmergeExecutionStepFactory(BaseExecutionStepFactory):
    def __init__(
        self,
        emerge_execution,
        emerge_part: UploadEmergePartDefinition,
        part_number,
        large_file_id,
        large_file_upload_state,
        finished_parts=None,
    ):
        super().__init__(emerge_execution, emerge_part)
        self.part_number = part_number
        self.large_file_id = large_file_id
        self.large_file_upload_state = large_file_upload_state
        self.finished_parts = finished_parts or {}

    def create_copy_execution_step(self, copy_range):
        return CopyPartExecutionStep(
            self.emerge_execution,
            copy_range,
            self.part_number,
            self.large_file_id,
            self.large_file_upload_state,
        )

    def create_upload_execution_step(self, stream_opener, stream_length=None, stream_sha1=None):
        return UploadPartExecutionStep(
            self.emerge_execution,
            stream_opener,
            self.part_number,
            self.large_file_id,
            self.large_file_upload_state,
            stream_length=stream_length,
            stream_sha1=stream_sha1,
            finished_parts=self.finished_parts,
        )


class BaseExecutionStep(metaclass=ABCMeta):
    def __init__(self, emerge_execution: BaseEmergeExecution):
        self.emerge_execution = emerge_execution

    @abstractmethod
    def execute(self):
        pass


class CopyFileExecutionStep(BaseExecutionStep):
    def __init__(self, emerge_execution, copy_source_range):
        super().__init__(emerge_execution)
        self.copy_source_range = copy_source_range

    def execute(self):
        execution = self.emerge_execution
        # if content type is not None then we support empty dict as default file info
        # but if content type is None, then setting empty dict as file info
        # would result with an error, because default in such case is: copy from source
        if execution.content_type is not None:
            file_info = execution.file_info or {}
        else:
            file_info = None
        return execution.services.copy_manager.copy_file(
            self.copy_source_range,
            execution.file_name,
            content_type=execution.content_type,
            file_info=file_info,
            destination_bucket_id=execution.bucket_id,
            progress_listener=execution.progress_listener,
            destination_encryption=execution.encryption,
            source_encryption=self.copy_source_range.encryption,
            file_retention=execution.file_retention,
            legal_hold=execution.legal_hold,
        )


class CopyPartExecutionStep(BaseExecutionStep):
    def __init__(
        self,
        emerge_execution,
        copy_source_range,
        part_number,
        large_file_id,
        large_file_upload_state,
        finished_parts=None,
    ):
        super().__init__(emerge_execution)
        self.copy_source_range = copy_source_range
        self.part_number = part_number
        self.large_file_id = large_file_id
        self.large_file_upload_state = large_file_upload_state
        self.finished_parts = finished_parts or {}

    def execute(self):
        return self.emerge_execution.services.copy_manager.copy_part(
            self.large_file_id,
            self.copy_source_range,
            self.part_number,
            self.large_file_upload_state,
            finished_parts=self.finished_parts,
            destination_encryption=self.emerge_execution.encryption,
            source_encryption=self.copy_source_range.encryption,
        )


class UploadFileExecutionStep(BaseExecutionStep):
    def __init__(self, emerge_execution, stream_opener, stream_length=None, stream_sha1=None):
        super().__init__(emerge_execution)
        self.stream_opener = stream_opener
        self.stream_length = stream_length
        self.stream_sha1 = stream_sha1

    def execute(self):
        upload_source = UploadSourceStream(
            self.stream_opener,
            stream_length=self.stream_length,
            stream_sha1=self.stream_sha1,
        )
        execution = self.emerge_execution
        return execution.services.upload_manager.upload_file(
            execution.bucket_id,
            upload_source,
            execution.file_name,
            execution.content_type or execution.DEFAULT_CONTENT_TYPE,
            execution.file_info or {},
            execution.progress_listener,
            encryption=execution.encryption,
            file_retention=execution.file_retention,
            legal_hold=execution.legal_hold,
            custom_upload_timestamp=execution.custom_upload_timestamp,
        )


class UploadPartExecutionStep(BaseExecutionStep):
    def __init__(
        self,
        emerge_execution,
        stream_opener,
        part_number,
        large_file_id,
        large_file_upload_state,
        stream_length=None,
        stream_sha1=None,
        finished_parts=None,
    ):
        super().__init__(emerge_execution)
        self.stream_opener = stream_opener
        self.stream_length = stream_length
        self.stream_sha1 = stream_sha1
        self.part_number = part_number
        self.large_file_id = large_file_id
        self.large_file_upload_state = large_file_upload_state
        self.finished_parts = finished_parts or {}

    def execute(self):
        execution = self.emerge_execution
        upload_source = UploadSourceStream(
            self.stream_opener,
            stream_length=self.stream_length,
            stream_sha1=self.stream_sha1,
        )
        return execution.services.upload_manager.upload_part(
            execution.bucket_id,
            self.large_file_id,
            upload_source,
            self.part_number,
            self.large_file_upload_state,
            finished_parts=self.finished_parts,
            encryption=execution.encryption,
        )



######################################################################
#
# File: b2sdk/_internal/sync/action.py
#
# Copyright 2019 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from __future__ import annotations

import contextlib
import functools
import logging
import os
from abc import ABCMeta, abstractmethod
from contextlib import suppress

from ..bucket import Bucket
from ..file_version import FileVersion
from ..http_constants import SRC_LAST_MODIFIED_MILLIS
from ..scan.path import B2Path
from ..sync.report import ProgressReport, SyncReport
from ..transfer.outbound.outbound_source import OutboundTransferSource
from ..transfer.outbound.upload_source import UploadSourceLocalFile
from ..utils.escape import escape_control_chars
from .encryption_provider import AbstractSyncEncryptionSettingsProvider
from .report import SyncFileReporter

logger = logging.getLogger(__name__)


class AbstractAction(metaclass=ABCMeta):
    """
    An action to take, such as uploading, downloading, or deleting
    a file.  Multi-threaded tasks create a sequence of Actions which
    are then run by a pool of threads.

    An action can depend on other actions completing.  An example of
    this is making sure a CreateBucketAction happens before an
    UploadFileAction.
    """

    def run(self, bucket: Bucket, reporter: ProgressReport, dry_run: bool = False):
        """
        Main action routine.

        :param bucket: a Bucket object
        :type bucket: b2sdk._internal.bucket.Bucket
        :param reporter: a place to report errors
        :param dry_run: if True, perform a dry run
        :type dry_run: bool
        """
        try:
            if not dry_run:
                self.do_action(bucket, reporter)
            self.do_report(bucket, reporter)
        except Exception as e:
            logger.exception('an exception occurred in a sync action')
            reporter.error(str(self) + ': ' + repr(e) + ' ' + str(e))
            raise  # Re-throw so we can identify failed actions

    @abstractmethod
    def get_bytes(self) -> int:
        """
        Return the number of bytes to transfer for this action.
        """

    @abstractmethod
    def do_action(self, bucket: Bucket, reporter: ProgressReport) -> None:
        """
        Perform the action, returning only after the action is completed.

        :param bucket: a Bucket object
        :param reporter: a place to report errors
        """

    @abstractmethod
    def do_report(self, bucket: Bucket, reporter: ProgressReport) -> None:
        """
        Report the action performed.

        :param bucket: a Bucket object
        :param reporter: a place to report errors
        """


class B2UploadAction(AbstractAction):
    """
    File uploading action.
    """

    def __init__(
        self,
        local_full_path: str,
        relative_name: str,
        b2_file_name: str,
        mod_time_millis: int,
        size: int,
        encryption_settings_provider: AbstractSyncEncryptionSettingsProvider,
    ):
        """
        :param local_full_path: a local file path
        :param relative_name: a relative file name
        :param b2_file_name: a name of a new remote file
        :param mod_time_millis: file modification time in milliseconds
        :param size: a file size
        :param encryption_settings_provider: encryption setting provider
        """
        self.local_full_path = local_full_path
        self.relative_name = relative_name
        self.b2_file_name = b2_file_name
        self.mod_time_millis = mod_time_millis
        self.size = size
        self.encryption_settings_provider = encryption_settings_provider
        self.large_file_sha1 = None

    def get_bytes(self) -> int:
        """
        Return file size.
        """
        return self.size

    @functools.cached_property
    def _upload_source(self) -> UploadSourceLocalFile:
        """Upload source if the file was to be uploaded in full"""
        # NOTE: We're caching this to ensure that sha1 is not recalculated.
        return UploadSourceLocalFile(self.local_full_path)

    def get_all_sources(self) -> list[OutboundTransferSource]:
        """Get list of sources required to complete this upload"""
        return [self._upload_source]

    def do_action(self, bucket: Bucket, reporter: ProgressReport) -> None:
        """
        Perform the uploading action, returning only after the action is completed.

        :param bucket: a Bucket object
        :param reporter: a place to report errors
        """
        if reporter:
            progress_listener = SyncFileReporter(reporter)
        else:
            progress_listener = None
        file_info = {SRC_LAST_MODIFIED_MILLIS: str(self.mod_time_millis)}
        encryption = self.encryption_settings_provider.get_setting_for_upload(
            bucket=bucket,
            b2_file_name=self.b2_file_name,
            file_info=file_info,
            length=self.size,
        )

        sources = self.get_all_sources()
        large_file_sha1 = None

        if len(sources) > 1:
            # The upload will be incremental, calculate the large_file_sha1
            large_file_sha1 = self._upload_source.get_content_sha1()

        with contextlib.ExitStack() as exit_stack:
            if progress_listener:
                exit_stack.enter_context(progress_listener)
            bucket.concatenate(
                sources,
                self.b2_file_name,
                progress_listener=progress_listener,
                file_info=file_info,
                encryption=encryption,
                large_file_sha1=large_file_sha1,
            )

    def do_report(self, bucket: Bucket, reporter: ProgressReport) -> None:
        """
        Report the uploading action performed.

        :param bucket: a Bucket object
        :param reporter: a place to report errors
        """
        reporter.print_completion(f'upload {escape_control_chars(self.relative_name)}')

    def __str__(self) -> str:
        return f'b2_upload({self.local_full_path}, {self.b2_file_name}, {self.mod_time_millis})'


class B2IncrementalUploadAction(B2UploadAction):
    def __init__(
        self,
        local_full_path: str,
        relative_name: str,
        b2_file_name: str,
        mod_time_millis: int,
        size: int,
        encryption_settings_provider: AbstractSyncEncryptionSettingsProvider,
        file_version: FileVersion | None = None,
        absolute_minimum_part_size: int | None = None,
    ):
        """
        :param local_full_path: a local file path
        :param relative_name: a relative file name
        :param b2_file_name: a name of a new remote file
        :param mod_time_millis: file modification time in milliseconds
        :param size: a file size
        :param encryption_settings_provider: encryption setting provider
        :param file_version: version of file currently on the server
        :param absolute_minimum_part_size: minimum file part size for large files
        """
        super().__init__(
            local_full_path,
            relative_name,
            b2_file_name,
            mod_time_millis,
            size,
            encryption_settings_provider,
        )
        self.file_version = file_version
        self.absolute_minimum_part_size = absolute_minimum_part_size

    def get_all_sources(self) -> list[OutboundTransferSource]:
        return self._upload_source.get_incremental_sources(
            self.file_version, self.absolute_minimum_part_size
        )


class B2HideAction(AbstractAction):
    def __init__(self, relative_name: str, b2_file_name: str):
        """
        :param relative_name: a relative file name
        :param b2_file_name: a name of a remote file
        """
        self.relative_name = relative_name
        self.b2_file_name = b2_file_name

    def get_bytes(self) -> int:
        """
        Return file size.

        :return: always zero
        :rtype: int
        """
        return 0

    def do_action(self, bucket: Bucket, reporter: ProgressReport) -> None:
        """
        Perform the hiding action, returning only after the action is completed.

        :param bucket: a Bucket object
        :param reporter: a place to report errors
        """
        bucket.hide_file(self.b2_file_name)

    def do_report(self, bucket: Bucket, reporter: SyncReport):
        """
        Report the hiding action performed.

        :param bucket: a Bucket object
        :param reporter: a place to report errors
        """
        reporter.update_transfer(1, 0)
        reporter.print_completion(f'hide   {escape_control_chars(self.relative_name)}')

    def __str__(self) -> str:
        return f'b2_hide({self.b2_file_name})'


class B2DownloadAction(AbstractAction):
    def __init__(
        self,
        source_path: B2Path,
        b2_file_name: str,
        local_full_path: str,
        encryption_settings_provider: AbstractSyncEncryptionSettingsProvider,
    ):
        """
        :param source_path: the file to be downloaded
        :param b2_file_name: b2_file_name
        :param local_full_path: a local file path
        :param encryption_settings_provider: encryption setting provider
        """
        self.source_path = source_path
        self.b2_file_name = b2_file_name
        self.local_full_path = local_full_path
        self.encryption_settings_provider = encryption_settings_provider

    def get_bytes(self) -> int:
        """
        Return file size.
        """
        return self.source_path.size

    def _ensure_directory_existence(self) -> None:
        # TODO: this can fail to multiple reasons (e.g. path is a file, permissions etc).
        #   We could provide nice exceptions for it.
        parent_dir = os.path.dirname(self.local_full_path)
        if not os.path.isdir(parent_dir):
            with suppress(OSError):
                os.makedirs(parent_dir)
        if not os.path.isdir(parent_dir):
            raise Exception(f'could not create directory {parent_dir}')

    def do_action(self, bucket: Bucket, reporter: ProgressReport) -> None:
        """
        Perform the downloading action, returning only after the action is completed.

        :param bucket: a Bucket object
        :param reporter: a place to report errors
        """
        self._ensure_directory_existence()

        if reporter:
            progress_listener = SyncFileReporter(reporter)
        else:
            progress_listener = None

        # Download the file to a .tmp file
        download_path = self.local_full_path + '.b2.sync.tmp'

        encryption = self.encryption_settings_provider.get_setting_for_download(
            bucket=bucket,
            file_version=self.source_path.selected_version,
        )
        with contextlib.ExitStack() as exit_stack:
            if progress_listener:
                exit_stack.enter_context(progress_listener)
            downloaded_file = bucket.download_file_by_id(
                self.source_path.selected_version.id_,
                progress_listener=progress_listener,
                encryption=encryption,
            )
            downloaded_file.save_to(download_path)

        # Move the file into place
        with suppress(OSError):
            os.unlink(self.local_full_path)
        os.rename(download_path, self.local_full_path)

    def do_report(self, bucket: Bucket, reporter: ProgressReport) -> None:
        """
        Report the downloading action performed.

        :param bucket: a Bucket object
        :param reporter: a place to report errors
        """
        reporter.print_completion('dnload ' + self.source_path.relative_path)

    def __str__(self) -> str:
        return 'b2_download(%s, %s, %s, %d)' % (
            self.b2_file_name,
            self.source_path.selected_version.id_,
            self.local_full_path,
            self.source_path.mod_time,
        )


class B2CopyAction(AbstractAction):
    """
    File copying action.
    """

    def __init__(
        self,
        b2_file_name: str,
        source_path: B2Path,
        dest_b2_file_name,
        source_bucket: Bucket,
        destination_bucket: Bucket,
        encryption_settings_provider: AbstractSyncEncryptionSettingsProvider,
    ):
        """
        :param b2_file_name: a b2_file_name
        :param source_path: the file to be copied
        :param dest_b2_file_name: a name of a destination remote file
        :param source_bucket: bucket to copy from
        :param destination_bucket: bucket to copy to
        :param encryption_settings_provider: encryption setting provider
        """
        self.b2_file_name = b2_file_name
        self.source_path = source_path
        self.dest_b2_file_name = dest_b2_file_name
        self.encryption_settings_provider = encryption_settings_provider
        self.source_bucket = source_bucket
        self.destination_bucket = destination_bucket

    def get_bytes(self) -> int:
        """
        Return file size.
        """
        return self.source_path.size

    def do_action(self, bucket: Bucket, reporter: ProgressReport) -> None:
        """
        Perform the copying action, returning only after the action is completed.

        :param bucket: a Bucket object
        :param reporter: a place to report errors
        """
        if reporter:
            progress_listener = SyncFileReporter(reporter)
        else:
            progress_listener = None

        source_encryption = self.encryption_settings_provider.get_source_setting_for_copy(
            bucket=self.source_bucket,
            source_file_version=self.source_path.selected_version,
        )

        destination_encryption = self.encryption_settings_provider.get_destination_setting_for_copy(
            bucket=self.destination_bucket,
            source_file_version=self.source_path.selected_version,
            dest_b2_file_name=self.dest_b2_file_name,
        )

        with contextlib.ExitStack() as exit_stack:
            if progress_listener:
                exit_stack.enter_context(progress_listener)
            bucket.copy(
                self.source_path.selected_version.id_,
                self.dest_b2_file_name,
                length=self.source_path.size,
                progress_listener=progress_listener,
                destination_encryption=destination_encryption,
                source_encryption=source_encryption,
                source_file_info=self.source_path.selected_version.file_info,
                source_content_type=self.source_path.selected_version.content_type,
            )

    def do_report(self, bucket: Bucket, reporter: ProgressReport) -> None:
        """
        Report the copying action performed.

        :param bucket: a Bucket object
        :param reporter: a place to report errors
        """
        reporter.print_completion('copy ' + self.source_path.relative_path)

    def __str__(self) -> str:
        return 'b2_copy(%s, %s, %s, %d)' % (
            self.b2_file_name,
            self.source_path.selected_version.id_,
            self.dest_b2_file_name,
            self.source_path.mod_time,
        )


class B2DeleteAction(AbstractAction):
    def __init__(self, relative_name: str, b2_file_name: str, file_id: str, note: str):
        """
        :param relative_name: a relative file name
        :param b2_file_name: a name of a remote file
        :param file_id: a file ID
        :param note: a deletion note
        """
        self.relative_name = relative_name
        self.b2_file_name = b2_file_name
        self.file_id = file_id
        self.note = note

    def get_bytes(self) -> int:
        """
        Return file size.

        :return: always zero
        """
        return 0

    def do_action(self, bucket: Bucket, reporter: ProgressReport):
        """
        Perform the deleting action, returning only after the action is completed.

        :param bucket: a Bucket object
        :param reporter: a place to report errors
        """
        bucket.api.delete_file_version(self.file_id, self.b2_file_name)

    def do_report(self, bucket: Bucket, reporter: SyncReport):
        """
        Report the deleting action performed.

        :param bucket: a Bucket object
        :param reporter: a place to report errors
        """
        reporter.update_transfer(1, 0)
        reporter.print_completion(f'delete {escape_control_chars(self.relative_name)} {self.note}')

    def __str__(self) -> str:
        return f'b2_delete({self.b2_file_name}, {self.file_id}, {self.note})'


class LocalDeleteAction(AbstractAction):
    def __init__(self, relative_name: str, full_path: str):
        """
        :param relative_name: a relative file name
        :param full_path: a full local path
        """
        self.relative_name = relative_name
        self.full_path = full_path

    def get_bytes(self) -> int:
        """
        Return file size.

        :return: always zero
        """
        return 0

    def do_action(self, bucket: Bucket, reporter: ProgressReport):
        """
        Perform the deleting of a local file action,
        returning only after the action is completed.

        :param bucket: a Bucket object
        :param reporter: a place to report errors
        """
        os.unlink(self.full_path)

    def do_report(self, bucket: Bucket, reporter: SyncReport):
        """
        Report the deleting of a local file action performed.

        :param bucket: a Bucket object
        :param reporter: a place to report errors
        """
        reporter.update_transfer(1, 0)
        reporter.print_completion(f'delete {escape_control_chars(self.relative_name)}')

    def __str__(self) -> str:
        return f'local_delete({self.full_path})'



######################################################################
#
# File: b2sdk/v2/utils.py
#
# Copyright 2022 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################

from __future__ import annotations

import shutil
import tempfile
import warnings


class TempDir:
    """
    Context manager that creates and destroys a temporary directory.
    """

    def __enter__(self):
        """
        Return the unicode path to the temp dir.
        """
        warnings.warn(
            'TempDir is deprecated. Use tempfile.TemporaryDirectory or pytest tmp_path fixture instead.',
            DeprecationWarning,
            stacklevel=2,
        )
        dirpath_bytes = tempfile.mkdtemp()
        self.dirpath = str(dirpath_bytes.replace('\\', '\\\\'))
        return self.dirpath

    def __exit__(self, exc_type, exc_val, exc_tb):
        shutil.rmtree(self.dirpath)
        return None  # do not hide exception



######################################################################
#
# File: b2sdk/v2/__init__.py
#
# Copyright 2021 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from __future__ import annotations

from b2sdk.v3 import *  # noqa
from b2sdk.v3 import parse_folder as parse_sync_folder
from b2sdk.v3 import AbstractPath as AbstractSyncPath
from b2sdk.v3 import LocalPath as LocalSyncPath
from b2sdk._internal.utils.escape import (
    unprintable_to_hex,
    escape_control_chars,
    substitute_control_chars,
)

from .account_info import AbstractAccountInfo
from .account_info import InMemoryAccountInfo
from .account_info import SqliteAccountInfo
from .account_info import StubAccountInfo
from .account_info import UrlPoolAccountInfo
from .api import B2Api, Services
from .b2http import B2Http
from .api_config import B2HttpApiConfig, DEFAULT_HTTP_API_CONFIG
from .bucket import Bucket, BucketFactory
from .session import B2Session
from .sync import B2SyncPath
from .transfer import DownloadManager, UploadManager

# replication

from .replication.setup import ReplicationSetupHelper

# data classes

from .application_key import ApplicationKey
from .application_key import BaseApplicationKey
from .application_key import FullApplicationKey

# utils

from .version_utils import rename_argument, rename_function
from .utils import TempDir

# raw_simulator

from .raw_simulator import BucketSimulator
from .raw_simulator import RawSimulator
from .raw_simulator import KeySimulator

# raw_api

from .raw_api import AbstractRawApi
from .raw_api import B2RawHTTPApi

# file_version

from .file_version import FileVersion
from .file_version import FileVersionFactory

# large_file

from .large_file import LargeFileServices
from .large_file import UnfinishedLargeFile



######################################################################
#
# File: b2sdk/v2/raw_api.py
#
# Copyright 2023 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from __future__ import annotations
from abc import abstractmethod

from b2sdk import v3
from b2sdk.v2._compat import _file_infos_rename

API_VERSION = 'v3'


class _OldRawAPI:
    """RawAPI compatibility layer"""

    @classmethod
    @_file_infos_rename
    def get_upload_file_headers(
        cls,
        upload_auth_token: str,
        file_name: str,
        content_length: int,
        content_type: str,
        content_sha1: str,
        file_info: dict,
        server_side_encryption: v3.EncryptionSetting | None,
        file_retention: v3.FileRetentionSetting | None,
        legal_hold: v3.LegalHold | None,
        custom_upload_timestamp: int | None = None,
        cache_control: str | None = None,
        *args,
        **kwargs,
    ) -> dict:
        if cache_control is not None:
            file_info['b2-cache-control'] = cache_control
        return super().get_upload_file_headers(
            upload_auth_token,
            file_name,
            content_length,
            content_type,
            content_sha1,
            file_info,
            server_side_encryption,
            file_retention,
            legal_hold,
            custom_upload_timestamp,
            *args,
            **kwargs,
        )

    def unprintable_to_hex(self, s):
        return v3.unprintable_to_hex(s)

    @_file_infos_rename
    def upload_file(
        self,
        upload_url,
        upload_auth_token,
        file_name,
        content_length,
        content_type,
        content_sha1,
        file_info: dict,
        data_stream,
        server_side_encryption: v3.EncryptionSetting | None = None,
        file_retention: v3.FileRetentionSetting | None = None,
        legal_hold: v3.LegalHold | None = None,
        custom_upload_timestamp: int | None = None,
        cache_control: str | None = None,
        *args,
        **kwargs,
    ):
        if cache_control is not None:
            file_info['b2-cache-control'] = cache_control
        return super().upload_file(
            upload_url,
            upload_auth_token,
            file_name,
            content_length,
            content_type,
            content_sha1,
            file_info,
            data_stream,
            server_side_encryption,
            file_retention,
            legal_hold,
            custom_upload_timestamp,
            *args,
            **kwargs,
        )

    def get_download_url_by_id(self, download_url, file_id):
        return f'{download_url}/b2api/{API_VERSION}/b2_download_file_by_id?fileId={file_id}'


class AbstractRawApi(_OldRawAPI, v3.AbstractRawApi):
    @abstractmethod
    def create_key(
        self,
        api_url,
        account_auth_token,
        account_id,
        capabilities,
        key_name,
        valid_duration_seconds,
        bucket_id,
        name_prefix,
    ):
        pass


class B2RawHTTPApi(_OldRawAPI, v3.B2RawHTTPApi):
    API_VERSION = API_VERSION

    def create_key(
        self,
        api_url,
        account_auth_token,
        account_id,
        capabilities,
        key_name,
        valid_duration_seconds,
        bucket_id,
        name_prefix,
    ):
        return self._post_json(
            api_url,
            'b2_create_key',
            account_auth_token,
            accountId=account_id,
            capabilities=capabilities,
            keyName=key_name,
            validDurationInSeconds=valid_duration_seconds,
            bucketId=bucket_id,
            namePrefix=name_prefix,
        )
